#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for multiscene.py."""

import os
import shutil
import tempfile
import unittest
from datetime import datetime
from unittest import mock

import pytest

from satpy.dataset.dataid import DataID, ModifierTuple, WavelengthRange

DEFAULT_SHAPE = (5, 10)

local_id_keys_config = {'name': {
    'required': True,
},
    'wavelength': {
    'type': WavelengthRange,
},
    'resolution': None,
    'calibration': {
    'enum': [
        'reflectance',
        'brightness_temperature',
        'radiance',
        'counts'
    ]
},
    'polarization': None,
    'level': None,
    'modifiers': {
    'required': True,
    'default': ModifierTuple(),
    'type': ModifierTuple,
},
}


def make_dataid(**items):
    """Make a data id."""
    return DataID(local_id_keys_config, **items)


def _fake_get_enhanced_image(img, enhance=None, overlay=None, decorate=None):
    from trollimage.xrimage import XRImage
    return XRImage(img)


def _create_test_area(proj_str=None, shape=DEFAULT_SHAPE, extents=None):
    """Create a test area definition."""
    from pyresample.geometry import AreaDefinition
    from pyresample.utils import proj4_str_to_dict
    if proj_str is None:
        proj_str = '+proj=lcc +datum=WGS84 +ellps=WGS84 +lon_0=-95. ' \
                   '+lat_0=25 +lat_1=25 +units=m +no_defs'
    proj_dict = proj4_str_to_dict(proj_str)
    extents = extents or (-1000., -1500., 1000., 1500.)

    return AreaDefinition(
        'test',
        'test',
        'test',
        proj_dict,
        shape[1],
        shape[0],
        extents
    )


def _create_test_dataset(name, shape=DEFAULT_SHAPE, area=None):
    """Create a test DataArray object."""
    import dask.array as da
    import numpy as np
    import xarray as xr

    return xr.DataArray(
        da.zeros(shape, dtype=np.float32, chunks=shape), dims=('y', 'x'),
        attrs={'name': name, 'area': area, '_satpy_id_keys': local_id_keys_config})


def _create_test_scenes(num_scenes=2, shape=DEFAULT_SHAPE, area=None):
    """Create some test scenes for various test cases."""
    from satpy import Scene
    ds1 = _create_test_dataset('ds1', shape=shape, area=area)
    ds2 = _create_test_dataset('ds2', shape=shape, area=area)
    scenes = []
    for _ in range(num_scenes):
        scn = Scene()
        scn['ds1'] = ds1.copy()
        scn['ds2'] = ds2.copy()
        scenes.append(scn)
    return scenes


class TestMultiScene(unittest.TestCase):
    """Test basic functionality of MultiScene."""

    def test_init_empty(self):
        """Test creating a multiscene with no children."""
        from satpy import MultiScene
        MultiScene()

    def test_init_children(self):
        """Test creating a multiscene with children."""
        from satpy import MultiScene
        scenes = _create_test_scenes()
        MultiScene(scenes)

    def test_properties(self):
        """Test basic properties/attributes of the MultiScene."""
        from satpy import MultiScene

        area = _create_test_area()
        scenes = _create_test_scenes(area=area)
        ds1_id = make_dataid(name='ds1')
        ds2_id = make_dataid(name='ds2')
        ds3_id = make_dataid(name='ds3')
        ds4_id = make_dataid(name='ds4')

        # Add a dataset to only one of the Scenes
        scenes[1]['ds3'] = _create_test_dataset('ds3')
        mscn = MultiScene(scenes)

        self.assertSetEqual(mscn.loaded_dataset_ids,
                            {ds1_id, ds2_id, ds3_id})
        self.assertSetEqual(mscn.shared_dataset_ids, {ds1_id, ds2_id})
        self.assertTrue(mscn.all_same_area)

        bigger_area = _create_test_area(shape=(20, 40))
        scenes[0]['ds4'] = _create_test_dataset('ds4', shape=(20, 40),
                                                area=bigger_area)

        self.assertSetEqual(mscn.loaded_dataset_ids,
                            {ds1_id, ds2_id, ds3_id, ds4_id})
        self.assertSetEqual(mscn.shared_dataset_ids, {ds1_id, ds2_id})
        self.assertFalse(mscn.all_same_area)

    def test_from_files(self):
        """Test creating a multiscene from multiple files."""
        from satpy import MultiScene
        input_files_abi = [
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171502203_e20171171504576_c20171171505018.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171507203_e20171171509576_c20171171510018.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171512203_e20171171514576_c20171171515017.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171517203_e20171171519577_c20171171520019.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171522203_e20171171524576_c20171171525020.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171527203_e20171171529576_c20171171530017.nc",
            ]
        input_files_glm = [
            "OR_GLM-L2-GLMC-M3_G16_s20171171500000_e20171171501000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171501000_e20171171502000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171502000_e20171171503000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171503000_e20171171504000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171504000_e20171171505000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171505000_e20171171506000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171506000_e20171171507000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171507000_e20171171508000_c20380190314080.nc",
        ]
        with mock.patch('satpy.multiscene.Scene') as scn_mock:
            mscn = MultiScene.from_files(
                    input_files_abi,
                    reader='abi_l1b',
                    scene_kwargs={"reader_kwargs": {}})
            assert len(mscn.scenes) == 6
            calls = [mock.call(
                filenames={'abi_l1b': [in_file_abi]},
                reader_kwargs={})
                for in_file_abi in input_files_abi]
            scn_mock.assert_has_calls(calls)

            scn_mock.reset_mock()
            with pytest.warns(DeprecationWarning):
                mscn = MultiScene.from_files(
                        input_files_abi + input_files_glm,
                        reader=('abi_l1b', "glm_l2"),
                        group_keys=["start_time"],
                        ensure_all_readers=True,
                        time_threshold=30)
            assert len(mscn.scenes) == 2
            calls = [mock.call(
                filenames={'abi_l1b': [in_file_abi], 'glm_l2': [in_file_glm]})
                for (in_file_abi, in_file_glm) in
                zip(input_files_abi[0:2],
                    [input_files_glm[2]] + [input_files_glm[7]])]
            scn_mock.assert_has_calls(calls)
            scn_mock.reset_mock()
            mscn = MultiScene.from_files(
                    input_files_abi + input_files_glm,
                    reader=('abi_l1b', "glm_l2"),
                    group_keys=["start_time"],
                    ensure_all_readers=False,
                    time_threshold=30)
            assert len(mscn.scenes) == 12

    def test_group(self):
        """Test group."""
        from satpy import MultiScene, Scene

        ds1 = _create_test_dataset(name='ds1')
        ds2 = _create_test_dataset(name='ds2')
        ds3 = _create_test_dataset(name='ds3')
        ds4 = _create_test_dataset(name='ds4')
        scene1 = Scene()
        scene1['ds1'] = ds1
        scene1['ds2'] = ds2
        scene2 = Scene()
        scene2['ds3'] = ds3
        scene2['ds4'] = ds4

        multi_scene = MultiScene([scene1, scene2])
        groups = {make_dataid(name='odd', wavelength=(1, 2, 3)): ['ds1', 'ds3'],
                  make_dataid(name='even', wavelength=(2, 3, 4)): ['ds2', 'ds4']}
        multi_scene.group(groups)

        self.assertSetEqual(multi_scene.shared_dataset_ids, set(groups.keys()))

    def test_add_group_aliases(self):
        """Test adding group aliases."""
        import types

        import numpy as np
        import xarray as xr

        from satpy import Scene
        from satpy.multiscene import add_group_aliases

        # Define test scenes
        ds_id1 = make_dataid(name='ds1', wavelength=(10.7, 10.8, 10.9))
        ds_id2 = make_dataid(name='ds2', wavelength=(1.9, 2.0, 2.1))
        ds_id3 = make_dataid(name='ds3', wavelength=(10.8, 10.9, 11.0))
        ds_id31 = make_dataid(name='ds31', polarization='H')

        scene1 = Scene()
        scene1[ds_id1] = xr.DataArray([1])
        scene2 = Scene()
        scene2[ds_id2] = xr.DataArray([2])
        scene3 = Scene()
        scene3[ds_id3] = xr.DataArray([3])
        scene3[ds_id31] = xr.DataArray([4])
        scenes = [scene1, scene2, scene3]

        # Define groups
        g1 = make_dataid(name='g1', wavelength=(10, 11, 12))
        g2 = make_dataid(name='g2', wavelength=(1, 2, 3), polarization='V')
        groups = {g1: ['ds1', 'ds3'], g2: ['ds2']}

        # Test adding aliases
        with_aliases = add_group_aliases(iter(scenes), groups)
        self.assertIsInstance(with_aliases, types.GeneratorType)
        with_aliases = list(with_aliases)
        self.assertSetEqual(set(with_aliases[0].keys()), {g1, ds_id1})
        self.assertSetEqual(set(with_aliases[1].keys()), {g2, ds_id2})
        self.assertSetEqual(set(with_aliases[2].keys()), {g1, ds_id3, ds_id31})

        np.testing.assert_array_equal(with_aliases[0]['g1'].values, [1])
        np.testing.assert_array_equal(with_aliases[0]['ds1'].values, [1])
        np.testing.assert_array_equal(with_aliases[1]['g2'].values, [2])
        np.testing.assert_array_equal(with_aliases[1]['ds2'].values, [2])
        np.testing.assert_array_equal(with_aliases[2]['g1'].values, [3])
        np.testing.assert_array_equal(with_aliases[2]['ds3'].values, [3])
        np.testing.assert_array_equal(with_aliases[2]['ds31'].values, [4])

        # Make sure that modifying the result doesn't modify the original
        self.assertNotIn(g1, scene1)

        # Adding an alias for multiple datasets in one scene should fail
        gen = add_group_aliases([scene3], {g1: ['ds3', 'ds31']})
        self.assertRaises(ValueError, list, gen)


class TestMultiSceneSave(unittest.TestCase):
    """Test saving a MultiScene to various formats."""

    def setUp(self):
        """Create temporary directory to save files to."""
        self.base_dir = tempfile.mkdtemp()

    def tearDown(self):
        """Remove the temporary directory created for a test."""
        try:
            shutil.rmtree(self.base_dir, ignore_errors=True)
        except OSError:
            pass

    @mock.patch('satpy.multiscene.get_enhanced_image', _fake_get_enhanced_image)
    def test_save_mp4_distributed(self):
        """Save a series of fake scenes to an mp4 video."""
        from satpy import MultiScene
        area = _create_test_area()
        scenes = _create_test_scenes(area=area)

        # Add a dataset to only one of the Scenes
        scenes[1]['ds3'] = _create_test_dataset('ds3')
        # Add a start and end time
        for ds_id in ['ds1', 'ds2', 'ds3']:
            scenes[1][ds_id].attrs['start_time'] = datetime(2018, 1, 2)
            scenes[1][ds_id].attrs['end_time'] = datetime(2018, 1, 2, 12)
            if ds_id == 'ds3':
                continue
            scenes[0][ds_id].attrs['start_time'] = datetime(2018, 1, 1)
            scenes[0][ds_id].attrs['end_time'] = datetime(2018, 1, 1, 12)

        mscn = MultiScene(scenes)
        fn = os.path.join(
            self.base_dir,
            'test_save_mp4_{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}.mp4')
        writer_mock = mock.MagicMock()
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v.compute() for v in x)
        client_mock.gather.side_effect = lambda x: x
        with mock.patch('satpy.multiscene.imageio.get_writer') as get_writer:
            get_writer.return_value = writer_mock
            # force order of datasets by specifying them
            mscn.save_animation(fn, client=client_mock, datasets=['ds1', 'ds2', 'ds3'])

        # 2 saves for the first scene + 1 black frame
        # 3 for the second scene
        self.assertEqual(writer_mock.append_data.call_count, 3 + 3)
        filenames = [os.path.basename(args[0][0]) for args in get_writer.call_args_list]
        self.assertEqual(filenames[0], 'test_save_mp4_ds1_20180101_00_20180102_12.mp4')
        self.assertEqual(filenames[1], 'test_save_mp4_ds2_20180101_00_20180102_12.mp4')
        self.assertEqual(filenames[2], 'test_save_mp4_ds3_20180102_00_20180102_12.mp4')

        # Test no distributed client found
        mscn = MultiScene(scenes)
        fn = os.path.join(
            self.base_dir,
            'test_save_mp4_{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}.mp4')
        writer_mock = mock.MagicMock()
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v.compute() for v in x)
        client_mock.gather.side_effect = lambda x: x
        with mock.patch('satpy.multiscene.imageio.get_writer') as get_writer, \
                mock.patch('satpy.multiscene.get_client', mock.Mock(side_effect=ValueError("No client"))):
            get_writer.return_value = writer_mock
            # force order of datasets by specifying them
            mscn.save_animation(fn, datasets=['ds1', 'ds2', 'ds3'])

        # 2 saves for the first scene + 1 black frame
        # 3 for the second scene
        self.assertEqual(writer_mock.append_data.call_count, 3 + 3)
        filenames = [os.path.basename(args[0][0]) for args in get_writer.call_args_list]
        self.assertEqual(filenames[0], 'test_save_mp4_ds1_20180101_00_20180102_12.mp4')
        self.assertEqual(filenames[1], 'test_save_mp4_ds2_20180101_00_20180102_12.mp4')
        self.assertEqual(filenames[2], 'test_save_mp4_ds3_20180102_00_20180102_12.mp4')

    @mock.patch('satpy.multiscene.get_enhanced_image', _fake_get_enhanced_image)
    def test_save_mp4_no_distributed(self):
        """Save a series of fake scenes to an mp4 video when distributed isn't available."""
        from satpy import MultiScene
        area = _create_test_area()
        scenes = _create_test_scenes(area=area)

        # Add a dataset to only one of the Scenes
        scenes[1]['ds3'] = _create_test_dataset('ds3')
        # Add a start and end time
        for ds_id in ['ds1', 'ds2', 'ds3']:
            scenes[1][ds_id].attrs['start_time'] = datetime(2018, 1, 2)
            scenes[1][ds_id].attrs['end_time'] = datetime(2018, 1, 2, 12)
            if ds_id == 'ds3':
                continue
            scenes[0][ds_id].attrs['start_time'] = datetime(2018, 1, 1)
            scenes[0][ds_id].attrs['end_time'] = datetime(2018, 1, 1, 12)

        mscn = MultiScene(scenes)
        fn = os.path.join(
            self.base_dir,
            'test_save_mp4_{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}.mp4')
        writer_mock = mock.MagicMock()
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v.compute() for v in x)
        client_mock.gather.side_effect = lambda x: x
        with mock.patch('satpy.multiscene.imageio.get_writer') as get_writer, \
                mock.patch('satpy.multiscene.get_client', None):
            get_writer.return_value = writer_mock
            # force order of datasets by specifying them
            mscn.save_animation(fn, datasets=['ds1', 'ds2', 'ds3'])

        # 2 saves for the first scene + 1 black frame
        # 3 for the second scene
        self.assertEqual(writer_mock.append_data.call_count, 3 + 3)
        filenames = [os.path.basename(args[0][0]) for args in get_writer.call_args_list]
        self.assertEqual(filenames[0], 'test_save_mp4_ds1_20180101_00_20180102_12.mp4')
        self.assertEqual(filenames[1], 'test_save_mp4_ds2_20180101_00_20180102_12.mp4')
        self.assertEqual(filenames[2], 'test_save_mp4_ds3_20180102_00_20180102_12.mp4')

    @mock.patch('satpy.multiscene.get_enhanced_image', _fake_get_enhanced_image)
    def test_save_datasets_simple(self):
        """Save a series of fake scenes to an PNG images."""
        from satpy import MultiScene
        area = _create_test_area()
        scenes = _create_test_scenes(area=area)

        # Add a dataset to only one of the Scenes
        scenes[1]['ds3'] = _create_test_dataset('ds3')
        # Add a start and end time
        for ds_id in ['ds1', 'ds2', 'ds3']:
            scenes[1][ds_id].attrs['start_time'] = datetime(2018, 1, 2)
            scenes[1][ds_id].attrs['end_time'] = datetime(2018, 1, 2, 12)
            if ds_id == 'ds3':
                continue
            scenes[0][ds_id].attrs['start_time'] = datetime(2018, 1, 1)
            scenes[0][ds_id].attrs['end_time'] = datetime(2018, 1, 1, 12)

        mscn = MultiScene(scenes)
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v for v in x)
        client_mock.gather.side_effect = lambda x: x
        with mock.patch('satpy.multiscene.Scene.save_datasets') as save_datasets:
            save_datasets.return_value = [True]  # some arbitrary return value
            # force order of datasets by specifying them
            mscn.save_datasets(base_dir=self.base_dir, client=False, datasets=['ds1', 'ds2', 'ds3'],
                               writer='simple_image')

        # 2 for each scene
        self.assertEqual(save_datasets.call_count, 2)

    @mock.patch('satpy.multiscene.get_enhanced_image', _fake_get_enhanced_image)
    def test_save_datasets_distributed_delayed(self):
        """Test distributed save for writers returning delayed obejcts e.g. simple_image."""
        from dask.delayed import Delayed

        from satpy import MultiScene
        area = _create_test_area()
        scenes = _create_test_scenes(area=area)

        # Add a dataset to only one of the Scenes
        scenes[1]['ds3'] = _create_test_dataset('ds3')
        # Add a start and end time
        for ds_id in ['ds1', 'ds2', 'ds3']:
            scenes[1][ds_id].attrs['start_time'] = datetime(2018, 1, 2)
            scenes[1][ds_id].attrs['end_time'] = datetime(2018, 1, 2, 12)
            if ds_id == 'ds3':
                continue
            scenes[0][ds_id].attrs['start_time'] = datetime(2018, 1, 1)
            scenes[0][ds_id].attrs['end_time'] = datetime(2018, 1, 1, 12)

        mscn = MultiScene(scenes)
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v for v in x)
        client_mock.gather.side_effect = lambda x: x
        future_mock = mock.MagicMock()
        future_mock.__class__ = Delayed
        with mock.patch('satpy.multiscene.Scene.save_datasets') as save_datasets:
            save_datasets.return_value = [future_mock]  # some arbitrary return value
            # force order of datasets by specifying them
            mscn.save_datasets(base_dir=self.base_dir, client=client_mock, datasets=['ds1', 'ds2', 'ds3'],
                               writer='simple_image')

        # 2 for each scene
        self.assertEqual(save_datasets.call_count, 2)

    @mock.patch('satpy.multiscene.get_enhanced_image', _fake_get_enhanced_image)
    def test_save_datasets_distributed_source_target(self):
        """Test distributed save for writers returning sources and targets e.g. geotiff writer."""
        import dask.array as da

        from satpy import MultiScene
        area = _create_test_area()
        scenes = _create_test_scenes(area=area)

        # Add a dataset to only one of the Scenes
        scenes[1]['ds3'] = _create_test_dataset('ds3')
        # Add a start and end time
        for ds_id in ['ds1', 'ds2', 'ds3']:
            scenes[1][ds_id].attrs['start_time'] = datetime(2018, 1, 2)
            scenes[1][ds_id].attrs['end_time'] = datetime(2018, 1, 2, 12)
            if ds_id == 'ds3':
                continue
            scenes[0][ds_id].attrs['start_time'] = datetime(2018, 1, 1)
            scenes[0][ds_id].attrs['end_time'] = datetime(2018, 1, 1, 12)

        mscn = MultiScene(scenes)
        client_mock = mock.MagicMock()
        client_mock.compute.side_effect = lambda x: tuple(v for v in x)
        client_mock.gather.side_effect = lambda x: x
        source_mock = mock.MagicMock()
        source_mock.__class__ = da.Array
        target_mock = mock.MagicMock()
        with mock.patch('satpy.multiscene.Scene.save_datasets') as save_datasets:
            save_datasets.return_value = [(source_mock, target_mock)]  # some arbitrary return value
            # force order of datasets by specifying them
            with self.assertRaises(NotImplementedError):
                mscn.save_datasets(base_dir=self.base_dir, client=client_mock, datasets=['ds1', 'ds2', 'ds3'],
                                   writer='geotiff')

    def test_crop(self):
        """Test the crop method."""
        import numpy as np
        from pyresample.geometry import AreaDefinition
        from xarray import DataArray

        from satpy import MultiScene, Scene
        scene1 = Scene()
        area_extent = (-5570248.477339745, -5561247.267842293, 5567248.074173927,
                       5570248.477339745)
        proj_dict = {'a': 6378169.0, 'b': 6356583.8, 'h': 35785831.0,
                     'lon_0': 0.0, 'proj': 'geos', 'units': 'm'}
        x_size = 3712
        y_size = 3712
        area_def = AreaDefinition(
            'test', 'test', 'test',
            proj_dict,
            x_size,
            y_size,
            area_extent,
        )
        area_def2 = AreaDefinition(
            'test2', 'test2', 'test2', proj_dict,
            x_size // 2,
            y_size // 2,
            area_extent,
            )
        scene1["1"] = DataArray(np.zeros((y_size, x_size)))
        scene1["2"] = DataArray(np.zeros((y_size, x_size)), dims=('y', 'x'))
        scene1["3"] = DataArray(np.zeros((y_size, x_size)), dims=('y', 'x'),
                                attrs={'area': area_def})
        scene1["4"] = DataArray(np.zeros((y_size // 2, x_size // 2)), dims=('y', 'x'),
                                attrs={'area': area_def2})
        mscn = MultiScene([scene1])

        # by lon/lat bbox
        new_mscn = mscn.crop(ll_bbox=(-20., -5., 0, 0))
        new_scn1 = list(new_mscn.scenes)[0]
        self.assertIn('1', new_scn1)
        self.assertIn('2', new_scn1)
        self.assertIn('3', new_scn1)
        self.assertTupleEqual(new_scn1['1'].shape, (y_size, x_size))
        self.assertTupleEqual(new_scn1['2'].shape, (y_size, x_size))
        self.assertTupleEqual(new_scn1['3'].shape, (184, 714))
        self.assertTupleEqual(new_scn1['4'].shape, (92, 357))


class TestBlendFuncs(unittest.TestCase):
    """Test individual functions used for blending."""

    def setUp(self):
        """Set up test data."""
        from datetime import datetime

        import dask.array as da
        import xarray as xr
        from pyresample.geometry import AreaDefinition
        area = AreaDefinition('test', 'test', 'test',
                              {'proj': 'geos', 'lon_0': -95.5, 'h': 35786023.0},
                              2, 2, [-200, -200, 200, 200])
        ds1 = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=('y', 'x'),
                           attrs={'start_time': datetime(2018, 1, 1, 0, 0, 0), 'area': area})
        self.ds1 = ds1
        ds2 = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=('y', 'x'),
                           attrs={'start_time': datetime(2018, 1, 1, 1, 0, 0), 'area': area})
        self.ds2 = ds2
        ds3 = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=('y', 'time'),
                           attrs={'start_time': datetime(2018, 1, 1, 0, 0, 0), 'area': area})
        self.ds3 = ds3
        ds4 = xr.DataArray(da.zeros((2, 2), chunks=-1), dims=('y', 'time'),
                           attrs={'start_time': datetime(2018, 1, 1, 1, 0, 0), 'area': area})
        self.ds4 = ds4

    def test_stack(self):
        """Test the 'stack' function."""
        from satpy.multiscene import stack
        res = stack([self.ds1, self.ds2])
        self.assertTupleEqual(self.ds1.shape, res.shape)

    def test_timeseries(self):
        """Test the 'timeseries' function."""
        import xarray as xr

        from satpy.multiscene import timeseries
        res = timeseries([self.ds1, self.ds2])
        res2 = timeseries([self.ds3, self.ds4])
        self.assertIsInstance(res, xr.DataArray)
        self.assertIsInstance(res2, xr.DataArray)
        self.assertTupleEqual((2, self.ds1.shape[0], self.ds1.shape[1]), res.shape)
        self.assertTupleEqual((self.ds3.shape[0], self.ds3.shape[1]+self.ds4.shape[1]), res2.shape)


@mock.patch('satpy.multiscene.get_enhanced_image')
def test_save_mp4(smg, tmp_path):
    """Save a series of fake scenes to an mp4 video."""
    from satpy import MultiScene
    area = _create_test_area()
    scenes = _create_test_scenes(area=area)
    smg.side_effect = _fake_get_enhanced_image

    # Add a dataset to only one of the Scenes
    scenes[1]['ds3'] = _create_test_dataset('ds3')
    # Add a start and end time
    for ds_id in ['ds1', 'ds2', 'ds3']:
        scenes[1][ds_id].attrs['start_time'] = datetime(2018, 1, 2)
        scenes[1][ds_id].attrs['end_time'] = datetime(2018, 1, 2, 12)
        if ds_id == 'ds3':
            continue
        scenes[0][ds_id].attrs['start_time'] = datetime(2018, 1, 1)
        scenes[0][ds_id].attrs['end_time'] = datetime(2018, 1, 1, 12)

    mscn = MultiScene(scenes)
    fn = str(tmp_path /
             'test_save_mp4_{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}.mp4')
    writer_mock = mock.MagicMock()
    with mock.patch('satpy.multiscene.imageio.get_writer') as get_writer:
        get_writer.return_value = writer_mock
        # force order of datasets by specifying them
        mscn.save_animation(fn, datasets=['ds1', 'ds2', 'ds3'], client=False)

    # 2 saves for the first scene + 1 black frame
    # 3 for the second scene
    assert writer_mock.append_data.call_count == 3 + 3
    filenames = [os.path.basename(args[0][0]) for args in get_writer.call_args_list]
    assert filenames[0] == 'test_save_mp4_ds1_20180101_00_20180102_12.mp4'
    assert filenames[1] == 'test_save_mp4_ds2_20180101_00_20180102_12.mp4'
    assert filenames[2] == 'test_save_mp4_ds3_20180102_00_20180102_12.mp4'

    # make sure that not specifying datasets still saves all of them
    fn = str(tmp_path /
             'test_save_mp4_{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}.mp4')
    writer_mock = mock.MagicMock()
    with mock.patch('satpy.multiscene.imageio.get_writer') as get_writer:
        get_writer.return_value = writer_mock
        # force order of datasets by specifying them
        mscn.save_animation(fn, client=False)
    # the 'ds3' dataset isn't known to the first scene so it doesn't get saved
    # 2 for first scene, 2 for second scene
    assert writer_mock.append_data.call_count == 2 + 2
    assert "test_save_mp4_ds1_20180101_00_20180102_12.mp4" in filenames
    assert "test_save_mp4_ds2_20180101_00_20180102_12.mp4" in filenames
    assert "test_save_mp4_ds3_20180102_00_20180102_12.mp4" in filenames

    # test decorating and enhancing

    fn = str(tmp_path /
             'test-{name}_{start_time:%Y%m%d_%H}_{end_time:%Y%m%d_%H}-rich.mp4')
    writer_mock = mock.MagicMock()
    with mock.patch('satpy.multiscene.imageio.get_writer') as get_writer:
        get_writer.return_value = writer_mock
        mscn.save_animation(
                fn, client=False,
                enh_args={"decorate": {
                    "decorate": [{
                        "text": {
                            "txt":
                            "Test {start_time:%Y-%m-%d %H:%M} - "
                            "{end_time:%Y-%m-%d %H:%M}"}}]}})
    assert writer_mock.append_data.call_count == 2 + 2
    assert ("2018-01-02" in smg.call_args_list[-1][1]
            ["decorate"]["decorate"][0]["text"]["txt"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.sar-c_safe module."""

import unittest
import unittest.mock as mock
from enum import Enum
from io import BytesIO

import dask.array as da
import numpy as np
import xarray as xr

from satpy.dataset import DataQuery
from satpy.readers.sar_c_safe import SAFEXMLAnnotation, SAFEXMLCalibration, SAFEXMLNoise


class TestSAFEGRD(unittest.TestCase):
    """Test the SAFE GRD file handler."""

    @mock.patch('rasterio.open')
    def setUp(self, mocked_rio_open):
        """Set up the test case."""
        from satpy.readers.sar_c_safe import SAFEGRD
        filename_info = {'mission_id': 'S1A', 'dataset_name': 'foo', 'start_time': 0, 'end_time': 0,
                         'polarization': 'vv'}
        filetype_info = 'bla'
        self.noisefh = mock.MagicMock()
        self.noisefh.get_noise_correction.return_value = xr.DataArray(np.zeros((2, 2)), dims=['y', 'x'])
        self.calfh = mock.MagicMock()
        self.calfh.get_calibration_constant.return_value = 1
        self.calfh.get_calibration.return_value = xr.DataArray(np.ones((2, 2)), dims=['y', 'x'])
        self.annotationfh = mock.MagicMock()

        self.test_fh = SAFEGRD('S1A_IW_GRDH_1SDV_20190201T024655_20190201T024720_025730_02DC2A_AE07.SAFE/measurement/'
                               's1a-iw-grd-vv-20190201t024655-20190201t024720-025730-02dc2a-001.tiff',
                               filename_info, filetype_info, self.calfh, self.noisefh, self.annotationfh)
        self.mocked_rio_open = mocked_rio_open

    def test_instantiate(self):
        """Test initialization of file handlers."""
        assert(self.test_fh._polarization == 'vv')
        assert(self.test_fh.calibration == self.calfh)
        assert(self.test_fh.noise == self.noisefh)
        self.mocked_rio_open.assert_called()

    @mock.patch('rioxarray.open_rasterio')
    def test_read_calibrated_natural(self, mocked_rioxarray_open):
        """Test the calibration routines."""
        calibration = mock.MagicMock()
        calibration.name = "sigma_nought"
        mocked_rioxarray_open.return_value = xr.DataArray(da.from_array(np.array([[0, 1], [2, 3]])), dims=['y', 'x'])
        xarr = self.test_fh.get_dataset(DataQuery(name="measurement", polarization="vv",
                                                  calibration=calibration, quantity='natural'), info=dict())
        np.testing.assert_allclose(xarr, [[np.nan, 2], [5, 10]])

    @mock.patch('rioxarray.open_rasterio')
    def test_read_calibrated_dB(self, mocked_rioxarray_open):
        """Test the calibration routines."""
        calibration = mock.MagicMock()
        calibration.name = "sigma_nought"
        mocked_rioxarray_open.return_value = xr.DataArray(da.from_array(np.array([[0, 1], [2, 3]])), dims=['y', 'x'])
        xarr = self.test_fh.get_dataset(DataQuery(name="measurement", polarization="vv",
                                                  calibration=calibration, quantity='dB'), info=dict())
        np.testing.assert_allclose(xarr, [[np.nan, 3.0103], [6.9897, 10]])

    def test_read_lon_lats(self):
        """Test reading lons and lats."""

        class FakeGCP:

            def __init__(self, *args):
                self.row, self.col, self.x, self.y, self.z = args

        gcps = [FakeGCP(0, 0, 0, 0, 0),
                FakeGCP(0, 3, 1, 0, 0),
                FakeGCP(3, 0, 0, 1, 0),
                FakeGCP(3, 3, 1, 1, 0),
                FakeGCP(0, 7, 2, 0, 0),
                FakeGCP(3, 7, 2, 1, 0),
                FakeGCP(7, 7, 2, 2, 0),
                FakeGCP(7, 3, 1, 2, 0),
                FakeGCP(7, 0, 0, 2, 0),
                FakeGCP(0, 15, 3, 0, 0),
                FakeGCP(3, 15, 3, 1, 0),
                FakeGCP(7, 15, 3, 2, 0),
                FakeGCP(15, 15, 3, 3, 0),
                FakeGCP(15, 7, 2, 3, 0),
                FakeGCP(15, 3, 1, 3, 0),
                FakeGCP(15, 0, 0, 3, 0),
                ]

        crs = dict(init='epsg:4326')

        self.mocked_rio_open.return_value.gcps = [gcps, crs]
        self.mocked_rio_open.return_value.shape = [16, 16]

        query = DataQuery(name="longitude", polarization="vv")
        xarr = self.test_fh.get_dataset(query, info=dict())
        expected = np.array([[3.79492915e-16, 5.91666667e-01, 9.09722222e-01,
                              1.00000000e+00, 9.08333333e-01, 6.80555556e-01,
                              3.62500000e-01, 8.32667268e-17, -3.61111111e-01,
                              -6.75000000e-01, -8.95833333e-01, -9.77777778e-01,
                              -8.75000000e-01, -5.41666667e-01, 6.80555556e-02,
                              1.00000000e+00],
                             [1.19166667e+00, 1.32437500e+00, 1.36941964e+00,
                              1.34166667e+00, 1.25598214e+00, 1.12723214e+00,
                              9.70282738e-01, 8.00000000e-01, 6.31250000e-01,
                              4.78898810e-01, 3.57812500e-01, 2.82857143e-01,
                              2.68898810e-01, 3.30803571e-01, 4.83437500e-01,
                              7.41666667e-01],
                             [1.82638889e+00, 1.77596726e+00, 1.72667765e+00,
                              1.67757937e+00, 1.62773172e+00, 1.57619402e+00,
                              1.52202558e+00, 1.46428571e+00, 1.40203373e+00,
                              1.33432894e+00, 1.26023065e+00, 1.17879819e+00,
                              1.08909084e+00, 9.90167942e-01, 8.81088790e-01,
                              7.60912698e-01],
                             [2.00000000e+00, 1.99166667e+00, 1.99305556e+00,
                              2.00000000e+00, 2.00833333e+00, 2.01388889e+00,
                              2.01250000e+00, 2.00000000e+00, 1.97222222e+00,
                              1.92500000e+00, 1.85416667e+00, 1.75555556e+00,
                              1.62500000e+00, 1.45833333e+00, 1.25138889e+00,
                              1.00000000e+00],
                             [1.80833333e+00, 2.01669643e+00, 2.18011267e+00,
                              2.30119048e+00, 2.38253827e+00, 2.42676446e+00,
                              2.43647747e+00, 2.41428571e+00, 2.36279762e+00,
                              2.28462160e+00, 2.18236607e+00, 2.05863946e+00,
                              1.91605017e+00, 1.75720663e+00, 1.58471726e+00,
                              1.40119048e+00],
                             [1.34722222e+00, 1.89627976e+00, 2.29940830e+00,
                              2.57341270e+00, 2.73509779e+00, 2.80126842e+00,
                              2.78872945e+00, 2.71428571e+00, 2.59474206e+00,
                              2.44690334e+00, 2.28757440e+00, 2.13356009e+00,
                              2.00166525e+00, 1.90869473e+00, 1.87145337e+00,
                              1.90674603e+00],
                             [7.12500000e-01, 1.67563988e+00, 2.36250177e+00,
                              2.80892857e+00, 3.05076318e+00, 3.12384850e+00,
                              3.06402742e+00, 2.90714286e+00, 2.68903770e+00,
                              2.44555485e+00, 2.21253720e+00, 2.02582766e+00,
                              1.92126913e+00, 1.93470451e+00, 2.10197669e+00,
                              2.45892857e+00],
                             [5.55111512e-16, 1.40000000e+00, 2.38095238e+00,
                              3.00000000e+00, 3.31428571e+00, 3.38095238e+00,
                              3.25714286e+00, 3.00000000e+00, 2.66666667e+00,
                              2.31428571e+00, 2.00000000e+00, 1.78095238e+00,
                              1.71428571e+00, 1.85714286e+00, 2.26666667e+00,
                              3.00000000e+00],
                             [-6.94444444e-01, 1.11458333e+00, 2.36631944e+00,
                              3.13888889e+00, 3.51041667e+00, 3.55902778e+00,
                              3.36284722e+00, 3.00000000e+00, 2.54861111e+00,
                              2.08680556e+00, 1.69270833e+00, 1.44444444e+00,
                              1.42013889e+00, 1.69791667e+00, 2.35590278e+00,
                              3.47222222e+00],
                             [-1.27500000e+00, 8.64613095e-01, 2.33016227e+00,
                              3.21785714e+00, 3.62390731e+00, 3.64452239e+00,
                              3.37591199e+00, 2.91428571e+00, 2.35585317e+00,
                              1.79682398e+00, 1.33340774e+00, 1.06181406e+00,
                              1.07825255e+00, 1.47893282e+00, 2.36006448e+00,
                              3.81785714e+00],
                             [-1.64583333e+00, 6.95312500e-01, 2.28404018e+00,
                              3.22916667e+00, 3.63950893e+00, 3.62388393e+00,
                              3.29110863e+00, 2.75000000e+00, 2.10937500e+00,
                              1.47805060e+00, 9.64843750e-01, 6.78571429e-01,
                              7.28050595e-01, 1.22209821e+00, 2.26953125e+00,
                              3.97916667e+00],
                             [-1.71111111e+00, 6.51904762e-01, 2.23951247e+00,
                              3.16507937e+00, 3.54197279e+00, 3.48356009e+00,
                              3.10320862e+00, 2.51428571e+00, 1.83015873e+00,
                              1.16419501e+00, 6.29761905e-01, 3.40226757e-01,
                              4.08956916e-01, 9.49319728e-01, 2.07468254e+00,
                              3.89841270e+00],
                             [-1.37500000e+00, 7.79613095e-01, 2.20813846e+00,
                              3.01785714e+00, 3.31605017e+00, 3.20999858e+00,
                              2.80698342e+00, 2.21428571e+00, 1.53918651e+00,
                              8.88966837e-01, 3.70907738e-01, 9.22902494e-02,
                              1.60395408e-01, 6.82504252e-01, 1.76589782e+00,
                              3.51785714e+00],
                             [-5.41666667e-01, 1.12366071e+00, 2.20147747e+00,
                              2.77976190e+00, 2.94649235e+00, 2.78964711e+00,
                              2.39720451e+00, 1.85714286e+00, 1.25744048e+00,
                              6.86075680e-01, 2.31026786e-01, -1.97278912e-02,
                              2.17899660e-02, 4.43558673e-01, 1.33355655e+00,
                              2.77976190e+00],
                             [8.84722222e-01, 1.72927083e+00, 2.23108879e+00,
                              2.44305556e+00, 2.41805060e+00, 2.20895337e+00,
                              1.86864335e+00, 1.45000000e+00, 1.00590278e+00,
                              5.89231151e-01, 2.52864583e-01, 4.96825397e-02,
                              3.25644841e-02, 2.54389881e-01, 7.68038194e-01,
                              1.62638889e+00],
                             [3.00000000e+00, 2.64166667e+00, 2.30853175e+00,
                              2.00000000e+00, 1.71547619e+00, 1.45436508e+00,
                              1.21607143e+00, 1.00000000e+00, 8.05555556e-01,
                              6.32142857e-01, 4.79166667e-01, 3.46031746e-01,
                              2.32142857e-01, 1.36904762e-01, 5.97222222e-02,
                              0.00000000e+00]])
        np.testing.assert_allclose(xarr.values, expected)


annotation_xml = b"""<?xml version="1.0" encoding="UTF-8"?>
<product>
  <adsHeader>
    <missionId>S1B</missionId>
    <productType>GRD</productType>
    <polarisation>HH</polarisation>
    <mode>EW</mode>
    <swath>EW</swath>
    <startTime>2020-03-15T05:04:28.137817</startTime>
    <stopTime>2020-03-15T05:05:32.416171</stopTime>
    <absoluteOrbitNumber>20698</absoluteOrbitNumber>
    <missionDataTakeId>160707</missionDataTakeId>
    <imageNumber>001</imageNumber>
  </adsHeader>
  <imageAnnotation>
    <imageInformation>
      <productFirstLineUtcTime>2020-03-15T05:04:28.137817</productFirstLineUtcTime>
      <productLastLineUtcTime>2020-03-15T05:05:32.416171</productLastLineUtcTime>
      <ascendingNodeTime>2020-03-15T04:33:22.256260</ascendingNodeTime>
      <anchorTime>2020-03-15T05:04:28.320641</anchorTime>
      <productComposition>Slice</productComposition>
      <sliceNumber>1</sliceNumber>
      <sliceList count="3">
        <slice>
          <sliceNumber>1</sliceNumber>
          <sensingStartTime>2020-03-15T05:04:29.485847</sensingStartTime>
          <sensingStopTime>2020-03-15T05:05:36.317420</sensingStopTime>
        </slice>
        <slice>
          <sliceNumber>2</sliceNumber>
          <sensingStartTime>2020-03-15T05:05:30.253413</sensingStartTime>
          <sensingStopTime>2020-03-15T05:06:34.046608</sensingStopTime>
        </slice>
        <slice>
          <sliceNumber>3</sliceNumber>
          <sensingStartTime>2020-03-15T05:06:31.020979</sensingStartTime>
          <sensingStopTime>2020-03-15T05:07:31.775796</sensingStopTime>
        </slice>
      </sliceList>
      <slantRangeTime>4.955163637998161e-03</slantRangeTime>
      <pixelValue>Detected</pixelValue>
      <outputPixels>16 bit Unsigned Integer</outputPixels>
      <rangePixelSpacing>4.000000e+01</rangePixelSpacing>
      <azimuthPixelSpacing>4.000000e+01</azimuthPixelSpacing>
      <azimuthTimeInterval>5.998353361537205e-03</azimuthTimeInterval>
      <azimuthFrequency>3.425601970000000e+02</azimuthFrequency>
      <numberOfSamples>10</numberOfSamples>
      <numberOfLines>10</numberOfLines>
      <zeroDopMinusAcqTime>-1.366569000000000e+00</zeroDopMinusAcqTime>
      <incidenceAngleMidSwath>3.468272707039038e+01</incidenceAngleMidSwath>
      <imageStatistics>
        <outputDataMean>
          <re>4.873919e+02</re>
          <im>0.000000e+00</im>
        </outputDataMean>
        <outputDataStdDev>
          <re>2.451083e+02</re>
          <im>0.000000e+00</im>
        </outputDataStdDev>
      </imageStatistics>
    </imageInformation>
  </imageAnnotation>
  <geolocationGrid>
    <geolocationGridPointList count="4">
      <geolocationGridPoint>
        <azimuthTime>2018-02-12T03:24:58.493342</azimuthTime>
        <slantRangeTime>4.964462411376810e-03</slantRangeTime>
        <line>0</line>
        <pixel>0</pixel>
        <latitude>7.021017981690355e+01</latitude>
        <longitude>5.609684402205929e+01</longitude>
        <height>8.234046399593353e-04</height>
        <incidenceAngle>1.918318045731997e+01</incidenceAngle>
        <elevationAngle>1.720012646010728e+01</elevationAngle>
      </geolocationGridPoint>
      <geolocationGridPoint>
        <azimuthTime>2018-02-12T03:24:58.493342</azimuthTime>
        <slantRangeTime>4.964462411376810e-03</slantRangeTime>
        <line>0</line>
        <pixel>9</pixel>
        <latitude>7.021017981690355e+01</latitude>
        <longitude>5.609684402205929e+01</longitude>
        <height>8.234046399593353e-04</height>
        <incidenceAngle>1.918318045731997e+01</incidenceAngle>
        <elevationAngle>1.720012646010728e+01</elevationAngle>
      </geolocationGridPoint>
      <geolocationGridPoint>
        <azimuthTime>2018-02-12T03:24:58.493342</azimuthTime>
        <slantRangeTime>4.964462411376810e-03</slantRangeTime>
        <line>9</line>
        <pixel>0</pixel>
        <latitude>7.021017981690355e+01</latitude>
        <longitude>5.609684402205929e+01</longitude>
        <height>8.234046399593353e-04</height>
        <incidenceAngle>1.918318045731997e+01</incidenceAngle>
        <elevationAngle>1.720012646010728e+01</elevationAngle>
      </geolocationGridPoint>
      <geolocationGridPoint>
        <azimuthTime>2018-02-12T03:24:58.493342</azimuthTime>
        <slantRangeTime>4.964462411376810e-03</slantRangeTime>
        <line>9</line>
        <pixel>9</pixel>
        <latitude>7.021017981690355e+01</latitude>
        <longitude>5.609684402205929e+01</longitude>
        <height>8.234046399593353e-04</height>
        <incidenceAngle>1.918318045731997e+01</incidenceAngle>
        <elevationAngle>1.720012646010728e+01</elevationAngle>
      </geolocationGridPoint>
    </geolocationGridPointList>
  </geolocationGrid>
</product>
"""

noise_xml = b"""<?xml version="1.0" encoding="UTF-8"?>
<noise>
  <noiseRangeVectorList count="3">
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>0</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 4.00000e+00 6.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>5</line>
      <pixel count="6">0 2 4 7 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 4.00000e+00 7.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>9</line>
      <pixel count="6">0 2 5 7 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 5.00000e+00 7.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
  </noiseRangeVectorList>
  <noiseAzimuthVectorList count="8">
    <noiseAzimuthVector>
      <swath>IW1</swath>
      <firstAzimuthLine>0</firstAzimuthLine>
      <firstRangeSample>1</firstRangeSample>
      <lastAzimuthLine>1</lastAzimuthLine>
      <lastRangeSample>3</lastRangeSample>
      <line count="1">0</line>
      <noiseAzimuthLut count="1">1.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW1</swath>
      <firstAzimuthLine>2</firstAzimuthLine>
      <firstRangeSample>0</firstRangeSample>
      <lastAzimuthLine>9</lastAzimuthLine>
      <lastRangeSample>1</lastRangeSample>
      <line count="4">2 4 6 8</line>
      <noiseAzimuthLut count="4">2.000000e+00 2.000000e+00 2.000000e+00 2.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>2</firstAzimuthLine>
      <firstRangeSample>2</firstRangeSample>
      <lastAzimuthLine>4</lastAzimuthLine>
      <lastRangeSample>4</lastRangeSample>
      <line count="2">2 4</line>
      <noiseAzimuthLut count="2">3.000000e+00 3.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>2</firstAzimuthLine>
      <firstRangeSample>5</firstRangeSample>
      <lastAzimuthLine>4</lastAzimuthLine>
      <lastRangeSample>8</lastRangeSample>
      <line count="2">2 4</line>
      <noiseAzimuthLut count="2">4.000000e+00 4.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>5</firstAzimuthLine>
      <firstRangeSample>2</firstRangeSample>
      <lastAzimuthLine>7</lastAzimuthLine>
      <lastRangeSample>5</lastRangeSample>
      <line count="2">5 6</line>
      <noiseAzimuthLut count="2">5.000000e+00 5.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>5</firstAzimuthLine>
      <firstRangeSample>6</firstRangeSample>
      <lastAzimuthLine>7</lastAzimuthLine>
      <lastRangeSample>9</lastRangeSample>
      <line count="2">5 6</line>
      <noiseAzimuthLut count="2">6.000000e+00 6.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>8</firstAzimuthLine>
      <firstRangeSample>2</firstRangeSample>
      <lastAzimuthLine>9</lastAzimuthLine>
      <lastRangeSample>6</lastRangeSample>
      <line count="1">8</line>
      <noiseAzimuthLut count="1">7.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>8</firstAzimuthLine>
      <firstRangeSample>7</firstRangeSample>
      <lastAzimuthLine>9</lastAzimuthLine>
      <lastRangeSample>9</lastRangeSample>
      <line count="1">8</line>
      <noiseAzimuthLut count="1">8.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
  </noiseAzimuthVectorList>
</noise>
"""

noise_xml_with_holes = b"""<?xml version="1.0" encoding="UTF-8"?>
<noise>
  <noiseRangeVectorList count="3">
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>0</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 4.00000e+00 6.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>5</line>
      <pixel count="6">0 2 4 7 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 4.00000e+00 7.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
    <noiseRangeVector>
      <azimuthTime>2020-03-15T05:04:28.137817</azimuthTime>
      <line>9</line>
      <pixel count="6">0 2 5 7 8 9</pixel>
      <noiseRangeLut count="6">0.00000e+00 2.00000e+00 5.00000e+00 7.00000e+00 8.00000e+00 9.00000e+00</noiseRangeLut>
    </noiseRangeVector>
  </noiseRangeVectorList>
  <noiseAzimuthVectorList count="12">
    <noiseAzimuthVector>
      <swath>IW1</swath>
      <firstAzimuthLine>0</firstAzimuthLine>
      <firstRangeSample>3</firstRangeSample>
      <lastAzimuthLine>2</lastAzimuthLine>
      <lastRangeSample>5</lastRangeSample>
      <line count="1">0</line>
      <noiseAzimuthLut count="1">1.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW1</swath>
      <firstAzimuthLine>1</firstAzimuthLine>
      <firstRangeSample>0</firstRangeSample>
      <lastAzimuthLine>5</lastAzimuthLine>
      <lastRangeSample>1</lastRangeSample>
      <line count="4">2 4 5</line>
      <noiseAzimuthLut count="4">2.000000e+00 2.000000e+00 2.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>2</firstAzimuthLine>
      <firstRangeSample>8</firstRangeSample>
      <lastAzimuthLine>4</lastAzimuthLine>
      <lastRangeSample>9</lastRangeSample>
      <line count="2">2 4</line>
      <noiseAzimuthLut count="2">3.000000e+00 3.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>3</firstAzimuthLine>
      <firstRangeSample>2</firstRangeSample>
      <lastAzimuthLine>5</lastAzimuthLine>
      <lastRangeSample>3</lastRangeSample>
      <line count="2">3 5</line>
      <noiseAzimuthLut count="2">4.000000e+00 4.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>3</firstAzimuthLine>
      <firstRangeSample>4</firstRangeSample>
      <lastAzimuthLine>4</lastAzimuthLine>
      <lastRangeSample>5</lastRangeSample>
      <line count="2">3 4</line>
      <noiseAzimuthLut count="2">5.000000e+00 5.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>4</firstAzimuthLine>
      <firstRangeSample>6</firstRangeSample>
      <lastAzimuthLine>4</lastAzimuthLine>
      <lastRangeSample>7</lastRangeSample>
      <line count="2">4</line>
      <noiseAzimuthLut count="2">6.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>5</firstAzimuthLine>
      <firstRangeSample>4</firstRangeSample>
      <lastAzimuthLine>7</lastAzimuthLine>
      <lastRangeSample>6</lastRangeSample>
      <line count="1">5 7</line>
      <noiseAzimuthLut count="1">7.000000e+00 7.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>5</firstAzimuthLine>
      <firstRangeSample>7</firstRangeSample>
      <lastAzimuthLine>7</lastAzimuthLine>
      <lastRangeSample>9</lastRangeSample>
      <line count="1">6</line>
      <noiseAzimuthLut count="1">8.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>6</firstAzimuthLine>
      <firstRangeSample>0</firstRangeSample>
      <lastAzimuthLine>7</lastAzimuthLine>
      <lastRangeSample>3</lastRangeSample>
      <line count="2">6 7</line>
      <noiseAzimuthLut count="2">9.000000e+00 9.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>8</firstAzimuthLine>
      <firstRangeSample>0</firstRangeSample>
      <lastAzimuthLine>9</lastAzimuthLine>
      <lastRangeSample>0</lastRangeSample>
      <line count="2">8</line>
      <noiseAzimuthLut count="2">10.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW2</swath>
      <firstAzimuthLine>8</firstAzimuthLine>
      <firstRangeSample>2</firstRangeSample>
      <lastAzimuthLine>9</lastAzimuthLine>
      <lastRangeSample>3</lastRangeSample>
      <line count="1">8 9</line>
      <noiseAzimuthLut count="1">11.000000e+00 11.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
    <noiseAzimuthVector>
      <swath>IW3</swath>
      <firstAzimuthLine>8</firstAzimuthLine>
      <firstRangeSample>4</firstRangeSample>
      <lastAzimuthLine>8</lastAzimuthLine>
      <lastRangeSample>5</lastRangeSample>
      <line count="1">8</line>
      <noiseAzimuthLut count="1">12.000000e+00</noiseAzimuthLut>
    </noiseAzimuthVector>
  </noiseAzimuthVectorList>
</noise>
"""


calibration_xml = b"""<?xml version="1.0" encoding="UTF-8"?>
<calibration>
  <adsHeader>
    <missionId>S1A</missionId>
    <productType>GRD</productType>
    <polarisation>VV</polarisation>
    <mode>IW</mode>
    <swath>IW</swath>
    <startTime>2018-02-12T03:24:58.493726</startTime>
    <stopTime>2018-02-12T03:25:01.493726</stopTime>
    <absoluteOrbitNumber>20568</absoluteOrbitNumber>
    <missionDataTakeId>144162</missionDataTakeId>
    <imageNumber>001</imageNumber>
  </adsHeader>
  <calibrationInformation>
    <absoluteCalibrationConstant>1.000000e+00</absoluteCalibrationConstant>
  </calibrationInformation>
  <calibrationVectorList count="4">
    <calibrationVector>
      <azimuthTime>2018-02-12T03:24:58.493726</azimuthTime>
      <line>0</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <sigmaNought count="6">1.894274e+03 1.788593e+03 1.320240e+03 1.277968e+03 1.277968e+03 1.277968e+03</sigmaNought>
      <betaNought count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</betaNought>
      <gamma count="6">1.840695e+03 1.718649e+03 1.187203e+03 1.185249e+03 1.183303e+03 1.181365e+03</gamma>
      <dn count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</dn>
    </calibrationVector>
    <calibrationVector>
      <azimuthTime>2018-02-12T03:24:59.493726</azimuthTime>
      <line>3</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <sigmaNought count="6">1.894274e+03 1.788593e+03 1.320240e+03 1.277968e+03 1.277968e+03 1.277968e+03</sigmaNought>
      <betaNought count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</betaNought>
      <gamma count="6">1.840695e+03 1.718649e+03 1.187203e+03 1.185249e+03 1.183303e+03 1.181365e+03</gamma>
      <dn count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</dn>
    </calibrationVector>
    <calibrationVector>
      <azimuthTime>2018-02-12T03:25:00.493726</azimuthTime>
      <line>6</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <sigmaNought count="6">1.894274e+03 1.788593e+03 1.320240e+03 1.277968e+03 1.277968e+03 1.277968e+03</sigmaNought>
      <betaNought count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</betaNought>
      <gamma count="6">1.840695e+03 1.718649e+03 1.187203e+03 1.185249e+03 1.183303e+03 1.181365e+03</gamma>
      <dn count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</dn>
    </calibrationVector>
    <calibrationVector>
      <azimuthTime>2018-02-12T03:25:01.493726</azimuthTime>
      <line>9</line>
      <pixel count="6">0 2 4 6 8 9</pixel>
      <sigmaNought count="6">1.894274e+03 1.788593e+03 1.320240e+03 1.277968e+03 1.277968e+03 1.277968e+03</sigmaNought>
      <betaNought count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</betaNought>
      <gamma count="6">1.840695e+03 1.718649e+03 1.187203e+03 1.185249e+03 1.183303e+03 1.181365e+03</gamma>
      <dn count="6">1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03 1.0870e+03</dn>
    </calibrationVector>
  </calibrationVectorList>
</calibration>
"""


class TestSAFEXMLNoise(unittest.TestCase):
    """Test the SAFE XML Noise file handler."""

    def setUp(self):
        """Set up the test case."""
        filename_info = dict(start_time=None, end_time=None, polarization="vv")
        self.annotation_fh = SAFEXMLAnnotation(BytesIO(annotation_xml), filename_info, mock.MagicMock())
        self.noise_fh = SAFEXMLNoise(BytesIO(noise_xml), filename_info, mock.MagicMock(), self.annotation_fh)

        self.expected_azimuth_noise = np.array([[np.nan, 1, 1, 1, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                                                [np.nan, 1, 1, 1, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                                                [2, 2, 3, 3, 3, 4, 4, 4, 4, np.nan],
                                                [2, 2, 3, 3, 3, 4, 4, 4, 4, np.nan],
                                                [2, 2, 3, 3, 3, 4, 4, 4, 4, np.nan],
                                                [2, 2, 5, 5, 5, 5, 6, 6, 6, 6],
                                                [2, 2, 5, 5, 5, 5, 6, 6, 6, 6],
                                                [2, 2, 5, 5, 5, 5, 6, 6, 6, 6],
                                                [2, 2, 7, 7, 7, 7, 7, 8, 8, 8],
                                                [2, 2, 7, 7, 7, 7, 7, 8, 8, 8],
                                                ])

        self.expected_range_noise = np.array([[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                                              ])

        self.noise_fh_with_holes = SAFEXMLNoise(BytesIO(noise_xml_with_holes), filename_info, mock.MagicMock(),
                                                self.annotation_fh)
        self.expected_azimuth_noise_with_holes = np.array(
            [[np.nan, np.nan, np.nan, 1, 1, 1, np.nan, np.nan, np.nan, np.nan],
             [2, 2, np.nan, 1, 1, 1, np.nan, np.nan, np.nan, np.nan],
             [2, 2, np.nan, 1, 1, 1, np.nan, np.nan, 3, 3],
             [2, 2, 4, 4, 5, 5, np.nan, np.nan, 3, 3],
             [2, 2, 4, 4, 5, 5, 6, 6, 3, 3],
             [2, 2, 4, 4, 7, 7, 7, 8, 8, 8],
             [9, 9, 9, 9, 7, 7, 7, 8, 8, 8],
             [9, 9, 9, 9, 7, 7, 7, 8, 8, 8],
             [10, np.nan, 11, 11, 12, 12, np.nan, np.nan, np.nan, np.nan],
             [10, np.nan, 11, 11, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
             ])

    def test_azimuth_noise_array(self):
        """Test reading the azimuth-noise array."""
        res = self.noise_fh.azimuth_noise_reader.read_azimuth_noise_array()
        np.testing.assert_array_equal(res, self.expected_azimuth_noise)

    def test_azimuth_noise_array_with_holes(self):
        """Test reading the azimuth-noise array."""
        res = self.noise_fh_with_holes.azimuth_noise_reader.read_azimuth_noise_array()
        np.testing.assert_array_equal(res, self.expected_azimuth_noise_with_holes)

    def test_range_noise_array(self):
        """Test reading the range-noise array."""
        res = self.noise_fh.read_range_noise_array(chunks=5)
        np.testing.assert_allclose(res, self.expected_range_noise)

    def test_get_noise_dataset(self):
        """Test using get_dataset for the noise."""
        query = DataQuery(name="noise", polarization="vv")
        res = self.noise_fh.get_dataset(query, {})
        np.testing.assert_allclose(res, self.expected_azimuth_noise * self.expected_range_noise)

    def test_get_noise_dataset_has_right_chunk_size(self):
        """Test using get_dataset for the noise has right chunk size in result."""
        query = DataQuery(name="noise", polarization="vv")
        res = self.noise_fh.get_dataset(query, {}, chunks=3)
        assert res.data.chunksize == (3, 3)


class Calibration(Enum):
    """Calibration levels."""

    gamma = 1
    sigma_nought = 2
    beta_nought = 3
    dn = 4


class TestSAFEXMLCalibration(unittest.TestCase):
    """Test the SAFE XML Calibration file handler."""

    def setUp(self):
        """Set up the test case."""
        filename_info = dict(start_time=None, end_time=None, polarization="vv")
        self.annotation_fh = SAFEXMLAnnotation(BytesIO(annotation_xml), filename_info, mock.MagicMock())
        self.calibration_fh = SAFEXMLCalibration(BytesIO(calibration_xml),
                                                 filename_info,
                                                 mock.MagicMock(),
                                                 self.annotation_fh)

        self.expected_gamma = np.array([[1840.695, 1779.672, 1718.649, 1452.926, 1187.203, 1186.226,
                                         1185.249, 1184.276, 1183.303, 1181.365]]) * np.ones((10, 1))

    def test_dn_calibration_array(self):
        """Test reading the dn calibration array."""
        expected_dn = np.ones((10, 10)) * 1087
        res = self.calibration_fh.get_calibration(Calibration.dn, chunks=5)
        np.testing.assert_allclose(res, expected_dn)

    def test_beta_calibration_array(self):
        """Test reading the beta calibration array."""
        expected_beta = np.ones((10, 10)) * 1087
        res = self.calibration_fh.get_calibration(Calibration.beta_nought, chunks=5)
        np.testing.assert_allclose(res, expected_beta)

    def test_sigma_calibration_array(self):
        """Test reading the sigma calibration array."""
        expected_sigma = np.array([[1894.274, 1841.4335, 1788.593, 1554.4165, 1320.24, 1299.104,
                                    1277.968, 1277.968, 1277.968, 1277.968]]) * np.ones((10, 1))
        res = self.calibration_fh.get_calibration(Calibration.sigma_nought, chunks=5)
        np.testing.assert_allclose(res, expected_sigma)

    def test_gamma_calibration_array(self):
        """Test reading the gamma calibration array."""
        res = self.calibration_fh.get_calibration(Calibration.gamma, chunks=5)
        np.testing.assert_allclose(res, self.expected_gamma)

    def test_get_calibration_dataset(self):
        """Test using get_dataset for the calibration."""
        query = DataQuery(name="gamma", polarization="vv")
        res = self.calibration_fh.get_dataset(query, {})
        np.testing.assert_allclose(res, self.expected_gamma)

    def test_get_calibration_dataset_has_right_chunk_size(self):
        """Test using get_dataset for the calibration yields array with right chunksize."""
        query = DataQuery(name="gamma", polarization="vv")
        res = self.calibration_fh.get_dataset(query, {}, chunks=3)
        assert res.data.chunksize == (3, 3)
        np.testing.assert_allclose(res, self.expected_gamma)

    def test_get_calibration_constant(self):
        """Test getting the calibration constant."""
        query = DataQuery(name="calibration_constant", polarization="vv")
        res = self.calibration_fh.get_dataset(query, {})
        assert res == 1


class TestSAFEXMLAnnotation(unittest.TestCase):
    """Test the SAFE XML Annotation file handler."""

    def setUp(self):
        """Set up the test case."""
        filename_info = dict(start_time=None, end_time=None, polarization="vv")
        self.annotation_fh = SAFEXMLAnnotation(BytesIO(annotation_xml), filename_info, mock.MagicMock())

    def test_incidence_angle(self):
        """Test reading the incidence angle."""
        query = DataQuery(name="incidence_angle", polarization="vv")
        res = self.annotation_fh.get_dataset(query, {})
        np.testing.assert_allclose(res, 19.18318046)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the geotiff writer."""

import unittest
from unittest import mock

import numpy as np


class TestGeoTIFFWriter(unittest.TestCase):
    """Test the GeoTIFF Writer class."""

    def setUp(self):
        """Create temporary directory to save files to."""
        import tempfile
        self.base_dir = tempfile.mkdtemp()

    def tearDown(self):
        """Remove the temporary directory created for a test."""
        try:
            import shutil
            shutil.rmtree(self.base_dir, ignore_errors=True)
        except OSError:
            pass

    def _get_test_datasets(self):
        """Create a single test dataset."""
        from datetime import datetime

        import dask.array as da
        import xarray as xr
        ds1 = xr.DataArray(
            da.zeros((100, 200), chunks=50),
            dims=('y', 'x'),
            attrs={'name': 'test',
                   'start_time': datetime.utcnow()}
        )
        return [ds1]

    def test_init(self):
        """Test creating the writer with no arguments."""
        from satpy.writers.geotiff import GeoTIFFWriter
        GeoTIFFWriter()

    def test_simple_write(self):
        """Test basic writer operation."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = self._get_test_datasets()
        w = GeoTIFFWriter(base_dir=self.base_dir)
        w.save_datasets(datasets)

    def test_simple_delayed_write(self):
        """Test writing can be delayed."""
        import dask.array as da

        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = self._get_test_datasets()
        w = GeoTIFFWriter(base_dir=self.base_dir)
        # when we switch to rio_save on XRImage then this will be sources
        # and targets
        res = w.save_datasets(datasets, compute=False)
        # this will fail if rasterio isn't installed
        self.assertIsInstance(res, tuple)
        # two lists, sources and destinations
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0], list)
        self.assertIsInstance(res[1], list)
        self.assertIsInstance(res[0][0], da.Array)
        da.store(res[0], res[1])
        for target in res[1]:
            if hasattr(target, 'close'):
                target.close()

    def test_colormap_write(self):
        """Test writing an image with a colormap."""
        from trollimage.colormap import spectral
        from trollimage.xrimage import XRImage

        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = self._get_test_datasets()
        w = GeoTIFFWriter(base_dir=self.base_dir)
        # we'd have to customize enhancements to test this through
        # save_datasets. We'll use `save_image` as a workaround.
        img = XRImage(datasets[0])
        img.palettize(spectral)
        w.save_image(img, keep_palette=True)

    def test_float_write(self):
        """Test that geotiffs can be written as floats.

        NOTE: Does not actually check that the output is floats.

        """
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = self._get_test_datasets()
        w = GeoTIFFWriter(base_dir=self.base_dir,
                          enhance=False,
                          dtype=np.float32)
        w.save_datasets(datasets)

    def test_dtype_for_enhance_false(self):
        """Test that dtype of dataset is used if parameters enhance=False and dtype=None."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = self._get_test_datasets()
        w = GeoTIFFWriter(base_dir=self.base_dir, enhance=False)
        with mock.patch('satpy.writers.XRImage.save') as save_method:
            save_method.return_value = None
            w.save_datasets(datasets, compute=False)
            self.assertEqual(save_method.call_args[1]['dtype'], np.float64)

    def test_dtype_for_enhance_false_and_given_dtype(self):
        """Test that dtype of dataset is used if enhance=False and dtype=uint8."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = self._get_test_datasets()
        w = GeoTIFFWriter(base_dir=self.base_dir, enhance=False, dtype=np.uint8)
        with mock.patch('satpy.writers.XRImage.save') as save_method:
            save_method.return_value = None
            w.save_datasets(datasets, compute=False)
            self.assertEqual(save_method.call_args[1]['dtype'], np.uint8)

    def test_fill_value_from_config(self):
        """Test fill_value coming from the writer config."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = self._get_test_datasets()
        w = GeoTIFFWriter(base_dir=self.base_dir)
        w.info['fill_value'] = 128
        with mock.patch('satpy.writers.XRImage.save') as save_method:
            save_method.return_value = None
            w.save_datasets(datasets, compute=False)
            self.assertEqual(save_method.call_args[1]['fill_value'], 128)

    def test_tags(self):
        """Test tags being added."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = self._get_test_datasets()
        w = GeoTIFFWriter(tags={'test1': 1}, base_dir=self.base_dir)
        w.info['fill_value'] = 128
        with mock.patch('satpy.writers.XRImage.save') as save_method:
            save_method.return_value = None
            w.save_datasets(datasets, tags={'test2': 2}, compute=False)
            called_tags = save_method.call_args[1]['tags']
            self.assertDictEqual(called_tags, {'test1': 1, 'test2': 2})

    def test_scale_offset(self):
        """Test tags being added."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = self._get_test_datasets()
        w = GeoTIFFWriter(tags={'test1': 1}, base_dir=self.base_dir)
        w.info['fill_value'] = 128
        with mock.patch('satpy.writers.XRImage.save') as save_method:
            save_method.return_value = None
            w.save_datasets(datasets, tags={'test2': 2}, compute=False, include_scale_offset=True)
            called_include = save_method.call_args[1]['include_scale_offset_tags']
            self.assertTrue(called_include)

    def test_tiled_value_from_config(self):
        """Test tiled value coming from the writer config."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = self._get_test_datasets()
        w = GeoTIFFWriter(base_dir=self.base_dir)
        with mock.patch('satpy.writers.XRImage.save') as save_method:
            save_method.return_value = None
            w.save_datasets(datasets, compute=False)
            self.assertEqual(save_method.call_args[1]['tiled'], True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The abi_l1b reader tests package."""

import unittest
from unittest import mock

import numpy as np
import pytest
import xarray as xr

from satpy.tests.utils import make_dataid


class Test_NC_ABI_L1B_Base(unittest.TestCase):
    """Common setup for NC_ABI_L1B tests."""

    @mock.patch('satpy.readers.abi_base.xr')
    def setUp(self, xr_, rad=None):
        """Create a fake dataset using the given radiance data."""
        from satpy.readers.abi_l1b import NC_ABI_L1B

        x_image = xr.DataArray(0.)
        y_image = xr.DataArray(0.)
        time = xr.DataArray(0.)
        if rad is None:
            rad_data = (np.arange(10.).reshape((2, 5)) + 1.) * 50.
            rad_data = (rad_data + 1.) / 0.5
            rad_data = rad_data.astype(np.int16)
            rad = xr.DataArray(
                rad_data,
                dims=('y', 'x'),
                attrs={
                    'scale_factor': 0.5,
                    'add_offset': -1.,
                    '_FillValue': 1002,
                    'units': 'W m-2 um-1 sr-1',
                    'valid_range': (0, 4095),
                }
            )
        rad.coords['t'] = time
        rad.coords['x_image'] = x_image
        rad.coords['y_image'] = y_image
        x__ = xr.DataArray(
            range(5),
            attrs={'scale_factor': 2., 'add_offset': -1.},
            dims=('x',)
        )
        y__ = xr.DataArray(
            range(2),
            attrs={'scale_factor': -2., 'add_offset': 1.},
            dims=('y',)
        )
        proj = xr.DataArray(
            [],
            attrs={
                'semi_major_axis': 1.,
                'semi_minor_axis': 1.,
                'perspective_point_height': 1.,
                'longitude_of_projection_origin': -90.,
                'latitude_of_projection_origin': 0.,
                'sweep_angle_axis': u'x'
            }
        )
        fake_dataset = xr.Dataset(
            data_vars={
                'Rad': rad,
                'band_id': np.array(8),
                # 'x': x__,
                # 'y': y__,
                'x_image': x_image,
                'y_image': y_image,
                'goes_imager_projection': proj,
                'yaw_flip_flag': np.array([1]),
                "planck_fk1": np.array(13432.1),
                "planck_fk2": np.array(1497.61),
                "planck_bc1": np.array(0.09102),
                "planck_bc2": np.array(0.99971),
                "esun": np.array(2017),
                "nominal_satellite_subpoint_lat": np.array(0.0),
                "nominal_satellite_subpoint_lon": np.array(-89.5),
                "nominal_satellite_height": np.array(35786.02),
                "earth_sun_distance_anomaly_in_AU": np.array(0.99)
            },
            coords={
                't': rad.coords['t'],
                'x': x__,
                'y': y__,

            },
            attrs={
                "time_coverage_start": "2017-09-20T17:30:40.8Z",
                "time_coverage_end": "2017-09-20T17:41:17.5Z",
            },
        )
        xr_.open_dataset.return_value = fake_dataset
        self.reader = NC_ABI_L1B('filename',
                                 {'platform_shortname': 'G16', 'observation_type': 'Rad',
                                  'suffix': 'custom',
                                  'scene_abbr': 'C', 'scan_mode': 'M3'},
                                 {'filetype': 'info'})


class TestABIYAML:
    """Tests for the ABI L1b reader's YAML configuration."""

    @pytest.mark.parametrize(['channel', 'suffix'],
                             [("C{:02d}".format(num), suffix)
                              for num in range(1, 17)
                              for suffix in ('', '_test_suffix')])
    def test_file_patterns_match(self, channel, suffix):
        """Test that the configured file patterns work."""
        from satpy.readers import configs_for_reader, load_reader
        reader_configs = list(configs_for_reader('abi_l1b'))[0]
        reader = load_reader(reader_configs)
        fn1 = ("OR_ABI-L1b-RadM1-M3{}_G16_s20182541300210_e20182541300267"
               "_c20182541300308{}.nc").format(channel, suffix)
        loadables = reader.select_files_from_pathnames([fn1])
        assert len(loadables) == 1
        if not suffix and channel in ["C01", "C02", "C03", "C05"]:
            fn2 = ("OR_ABI-L1b-RadM1-M3{}_G16_s20182541300210_e20182541300267"
                   "_c20182541300308-000000_0.nc").format(channel)
            loadables = reader.select_files_from_pathnames([fn2])
            assert len(loadables) == 1


class Test_NC_ABI_L1B(Test_NC_ABI_L1B_Base):
    """Test the NC_ABI_L1B reader."""

    def test_basic_attributes(self):
        """Test getting basic file attributes."""
        from datetime import datetime
        self.assertEqual(self.reader.start_time,
                         datetime(2017, 9, 20, 17, 30, 40, 800000))
        self.assertEqual(self.reader.end_time,
                         datetime(2017, 9, 20, 17, 41, 17, 500000))

    def test_get_dataset(self):
        """Test the get_dataset method."""
        key = make_dataid(name='Rad', calibration='radiance')
        res = self.reader.get_dataset(key, {'info': 'info'})
        exp = {'calibration': 'radiance',
               'instrument_ID': None,
               'modifiers': (),
               'name': 'Rad',
               'observation_type': 'Rad',
               'orbital_parameters': {'projection_altitude': 1.0,
                                      'projection_latitude': 0.0,
                                      'projection_longitude': -90.0,
                                      'satellite_nominal_altitude': 35786020.,
                                      'satellite_nominal_latitude': 0.0,
                                      'satellite_nominal_longitude': -89.5,
                                      'yaw_flip': True},
               'orbital_slot': None,
               'platform_name': 'GOES-16',
               'platform_shortname': 'G16',
               'production_site': None,
               'scan_mode': 'M3',
               'scene_abbr': 'C',
               'scene_id': None,
               'sensor': 'abi',
               'timeline_ID': None,
               'suffix': 'custom',
               'units': 'W m-2 um-1 sr-1'}

        self.assertDictEqual(res.attrs, exp)
        # we remove any time dimension information
        self.assertNotIn('t', res.coords)
        self.assertNotIn('t', res.dims)
        self.assertNotIn('time', res.coords)
        self.assertNotIn('time', res.dims)

    @mock.patch('satpy.readers.abi_base.geometry.AreaDefinition')
    def test_get_area_def(self, adef):
        """Test the area generation."""
        self.reader.get_area_def(None)

        self.assertEqual(adef.call_count, 1)
        call_args = tuple(adef.call_args)[0]
        self.assertDictEqual(call_args[3], {'a': 1.0, 'b': 1.0, 'h': 1.0, 'lon_0': -90.0, 'proj': 'geos',
                                            'sweep': 'x', 'units': 'm'})
        self.assertEqual(call_args[4], self.reader.ncols)
        self.assertEqual(call_args[5], self.reader.nlines)
        np.testing.assert_allclose(call_args[6], (-2, -2, 8, 2))


class Test_NC_ABI_L1B_ir_cal(Test_NC_ABI_L1B_Base):
    """Test the NC_ABI_L1B reader's IR calibration."""

    def setUp(self):
        """Create fake data for the tests."""
        rad_data = (np.arange(10.).reshape((2, 5)) + 1.) * 50.
        rad_data = (rad_data + 1.) / 0.5
        rad_data = rad_data.astype(np.int16)
        rad = xr.DataArray(
            rad_data,
            dims=('y', 'x'),
            attrs={
                'scale_factor': 0.5,
                'add_offset': -1.,
                '_FillValue': 1002,
            }
        )
        super(Test_NC_ABI_L1B_ir_cal, self).setUp(rad=rad)

    def test_ir_calibrate(self):
        """Test IR calibration."""
        res = self.reader.get_dataset(
            make_dataid(name='C05', calibration='brightness_temperature'), {})

        expected = np.array([[267.55572248, 305.15576503, 332.37383249, 354.73895301, 374.19710115],
                             [391.68679226, 407.74064808, 422.69329105, 436.77021913, np.nan]])
        self.assertTrue(np.allclose(res.data, expected, equal_nan=True))
        # make sure the attributes from the file are in the data array
        self.assertNotIn('scale_factor', res.attrs)
        self.assertNotIn('_FillValue', res.attrs)
        self.assertEqual(res.attrs['standard_name'],
                         'toa_brightness_temperature')
        self.assertEqual(res.attrs['long_name'], 'Brightness Temperature')


class Test_NC_ABI_L1B_vis_cal(Test_NC_ABI_L1B_Base):
    """Test the NC_ABI_L1B reader."""

    def setUp(self):
        """Create fake data for the tests."""
        rad_data = (np.arange(10.).reshape((2, 5)) + 1.)
        rad_data = (rad_data + 1.) / 0.5
        rad_data = rad_data.astype(np.int16)
        rad = xr.DataArray(
            rad_data,
            dims=('y', 'x'),
            attrs={
                'scale_factor': 0.5,
                'add_offset': -1.,
                '_FillValue': 20,
            }
        )
        super(Test_NC_ABI_L1B_vis_cal, self).setUp(rad=rad)

    def test_vis_calibrate(self):
        """Test VIS calibration."""
        res = self.reader.get_dataset(
            make_dataid(name='C05', calibration='reflectance'), {})

        expected = np.array([[0.15265617, 0.30531234, 0.45796851, 0.61062468, 0.76328085],
                             [0.91593702, 1.06859319, 1.22124936, np.nan, 1.52656171]])
        self.assertTrue(np.allclose(res.data, expected, equal_nan=True))
        self.assertNotIn('scale_factor', res.attrs)
        self.assertNotIn('_FillValue', res.attrs)
        self.assertEqual(res.attrs['standard_name'],
                         'toa_bidirectional_reflectance')
        self.assertEqual(res.attrs['long_name'],
                         'Bidirectional Reflectance')


class Test_NC_ABI_L1B_raw_cal(Test_NC_ABI_L1B_Base):
    """Test the NC_ABI_L1B reader raw calibration."""

    def setUp(self):
        """Create fake data for the tests."""
        rad_data = (np.arange(10.).reshape((2, 5)) + 1.)
        rad_data = (rad_data + 1.) / 0.5
        rad_data = rad_data.astype(np.int16)
        rad = xr.DataArray(
            rad_data,
            dims=('y', 'x'),
            attrs={
                'scale_factor': 0.5,
                'add_offset': -1.,
                '_FillValue': 20,
            }
        )
        super(Test_NC_ABI_L1B_raw_cal, self).setUp(rad=rad)

    def test_raw_calibrate(self):
        """Test RAW calibration."""
        res = self.reader.get_dataset(
            make_dataid(name='C05', calibration='counts'), {})

        # We expect the raw data to be unchanged
        expected = res.data
        self.assertTrue(np.allclose(res.data, expected, equal_nan=True))

        # check for the presence of typical attributes
        self.assertIn('scale_factor', res.attrs)
        self.assertIn('add_offset', res.attrs)
        self.assertIn('_FillValue', res.attrs)
        self.assertIn('orbital_parameters', res.attrs)
        self.assertIn('platform_shortname', res.attrs)
        self.assertIn('scene_id', res.attrs)

        # determine if things match their expected values/types.
        self.assertEqual(res.data.dtype, np.int16, "int16 data type expected")
        self.assertEqual(res.attrs['standard_name'],
                         'counts')
        self.assertEqual(res.attrs['long_name'],
                         'Raw Counts')


class Test_NC_ABI_L1B_invalid_cal(Test_NC_ABI_L1B_Base):
    """Test the NC_ABI_L1B reader with invalid calibration."""

    def test_invalid_calibration(self):
        """Test detection of invalid calibration values."""
        # Need to use a custom DataID class because the real DataID class is
        # smart enough to detect the invalid calibration before the ABI L1B
        # get_dataset method gets a chance to run.
        class FakeDataID(dict):
            def to_dict(self):
                return self

        with self.assertRaises(ValueError, msg='Did not detect invalid cal'):
            did = FakeDataID(name='C05', calibration='invalid', modifiers=())
            self.reader.get_dataset(did, {})


class Test_NC_ABI_File(unittest.TestCase):
    """Test file opening."""

    @mock.patch('satpy.readers.abi_base.xr')
    def test_open_dataset(self, _):
        """Test openning a dataset."""
        from satpy.readers.abi_l1b import NC_ABI_L1B

        openable_thing = mock.MagicMock()

        NC_ABI_L1B(openable_thing, {'platform_shortname': 'g16'}, None)
        openable_thing.open.assert_called()


class Test_NC_ABI_L1B_H5netcdf(Test_NC_ABI_L1B):
    """Allow h5netcdf peculiarities."""

    def setUp(self):
        """Create fake data for the tests."""
        rad_data = np.int16(50)
        rad = xr.DataArray(
            rad_data,
            attrs={
                'scale_factor': 0.5,
                'add_offset': -1.,
                '_FillValue': np.array([1002]),
                'units': 'W m-2 um-1 sr-1',
                'valid_range': (0, 4095),
            }
        )
        super(Test_NC_ABI_L1B_H5netcdf, self).setUp(rad=rad)



# -*- coding: utf-8 -*-
# Copyright (c) 2020,2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""HY-2B L2B Reader.

Distributed by Eumetsat in HDF5 format.
Also handle the HDF5 files from NSOAS, based on a file example.
"""

from datetime import datetime

import numpy as np
import xarray as xr

from satpy.readers.hdf5_utils import HDF5FileHandler


class HY2SCATL2BH5FileHandler(HDF5FileHandler):
    """File handler for HY2 scat."""

    @property
    def start_time(self):
        """Time for first observation."""
        return datetime.strptime(self['/attr/Range_Beginning_Time'],
                                 '%Y%m%dT%H:%M:%S')

    @property
    def end_time(self):
        """Time for final observation."""
        return datetime.strptime(self['/attr/Range_Ending_Time'],
                                 '%Y%m%dT%H:%M:%S')

    @property
    def platform_name(self):
        """Get the Platform ShortName."""
        return self['/attr/Platform_ShortName']

    def get_variable_metadata(self):
        """Get the variable metadata."""
        info = getattr(self, 'attrs', {})
        info.update({
            "Equator_Crossing_Longitude": self['/attr/Equator_Crossing_Longitude'],
            "Equator_Crossing_Time": self['/attr/Equator_Crossing_Time'],
            "Input_L2A_Filename": self['/attr/Input_L2A_Filename'],
            "L2B_Actual_WVC_Rows": self['/attr/L2B_Actual_WVC_Rows'],
            "Orbit_Inclination": self['/attr/Orbit_Inclination'],
            "Orbit_Number": self['/attr/Orbit_Number'],
            "Output_L2B_Filename": self['/attr/Output_L2B_Filename'],
            "Production_Date_Time": self['/attr/Production_Date_Time'],
            "L2B_Expected_WVC_Rows": self['/attr/L2B_Expected_WVC_Rows']
        })
        try:
            info.update({"L2B_Number_WVC_cells": self['/attr/L2B_Number_WVC_cells']})
        except KeyError:
            info.update({"L2B_Expected_WVC_Cells": self['/attr/L2B_Expected_WVC_Cells']})
        return info

    def get_metadata(self):
        """Get the metadata."""
        info = getattr(self, 'attrs', {})
        info.update({
            "WVC_Size": self['/attr/WVC_Size'],
            "HDF_Version_Id": self['/attr/HDF_Version_Id'],
            "Instrument_ShorName": self['/attr/Instrument_ShorName'],
            "L2A_Inputdata_Version": self['/attr/L2A_Inputdata_Version'],
            "L2B_Algorithm_Descriptor": self['/attr/L2B_Algorithm_Descriptor'],
            "L2B_Data_Version": self['/attr/L2B_Data_Version'],
            "L2B_Processing_Type": self['/attr/L2B_Processing_Type'],
            "L2B_Processor_Name": self['/attr/L2B_Processor_Name'],
            "L2B_Processor_Version": self['/attr/L2B_Processor_Version'],
            "Long_Name": self['/attr/Long_Name'],
            "Platform_LongName": self['/attr/Platform_LongName'],
            "Platform_ShortName": self['/attr/Platform_ShortName'],
            "Platform_Type": self['/attr/Platform_Type'],
            "Producer_Agency": self['/attr/Producer_Agency'],
            "Producer_Institution": self['/attr/Producer_Institution'],
            "Rev_Orbit_Perio": self['/attr/Rev_Orbit_Period'],
            "Short_Name": self['/attr/Short_Name'],
            "Sigma0_Granularity": self['/attr/Sigma0_Granularity'],
        })
        return info

    def get_dataset(self, key, info):
        """Get the dataset."""
        dims = ['y', 'x']
        if self[key['name']].ndim == 3:
            dims = ['y', 'x', 'selection']
        data = self[key['name']]
        if "valid range" in data.attrs:
            data.attrs.update({'valid_range': data.attrs.pop('valid range')})
        if key['name'] in 'wvc_row_time':
            data = data.rename({data.dims[0]: 'y'})
        else:
            dim_map = {curr_dim: new_dim for curr_dim, new_dim in zip(data.dims, dims)}
            data = data.rename(dim_map)
            data = self._mask_data(key['name'], data)
            data = self._scale_data(key['name'], data)

            if key['name'] in 'wvc_lon':
                data = xr.where(data > 180, data - 360., data)
        data.attrs.update(info)
        data.attrs.update(self.get_metadata())
        data.attrs.update(self.get_variable_metadata())
        if "Platform_ShortName" in data.attrs:
            data.attrs.update({'platform_name': data.attrs['Platform_ShortName']})

        return data

    def _scale_data(self, key_name, data):
        return data * self[key_name].attrs['scale_factor'] + self[key_name].attrs['add_offset']

    def _mask_data(self, key_name, data):
        data = xr.where(data == self[key_name].attrs['fill_value'], np.nan, data)

        valid_range = self[key_name].attrs['valid_range']
        data = xr.where(data < valid_range[0], np.nan, data)
        data = xr.where(data > valid_range[1], np.nan, data)
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'seadas_l2' reader."""

import numpy as np
import pytest
from pyhdf.SD import SD, SDC
from pyresample.geometry import SwathDefinition
from pytest_lazyfixture import lazy_fixture

from satpy import Scene, available_readers


@pytest.fixture(scope="module")
def seadas_l2_modis_chlor_a(tmp_path_factory):
    """Create MODIS SEADAS file."""
    filename = "a1.21322.1758.seadas.hdf"
    full_path = str(tmp_path_factory.mktemp("seadas_l2") / filename)
    return _create_seadas_chlor_a_file(full_path, "Aqua", "MODISA")


@pytest.fixture(scope="module")
def seadas_l2_viirs_npp_chlor_a(tmp_path_factory):
    """Create VIIRS NPP SEADAS file."""
    filename = "SEADAS_npp_d20211118_t1728125_e1739327.hdf"
    full_path = str(tmp_path_factory.mktemp("seadas") / filename)
    return _create_seadas_chlor_a_file(full_path, "NPP", "VIIRSN")


@pytest.fixture(scope="module")
def seadas_l2_viirs_j01_chlor_a(tmp_path_factory):
    """Create VIIRS JPSS-01 SEADAS file."""
    filename = "SEADAS_j01_d20211118_t1728125_e1739327.hdf"
    full_path = str(tmp_path_factory.mktemp("seadas") / filename)
    return _create_seadas_chlor_a_file(full_path, "JPSS-1", "VIIRSJ1")


def _create_seadas_chlor_a_file(full_path, mission, sensor):
    h = SD(full_path, SDC.WRITE | SDC.CREATE)
    setattr(h, "Sensor Name", sensor)
    h.Mission = mission
    setattr(h, "Start Time", "2021322175853191")
    setattr(h, "End Time", "2021322180551214")

    lon_info = {
        "type": SDC.FLOAT32,
        "data": np.zeros((5, 5), dtype=np.float32),
        "dim_labels": ["Number of Scan Lines", "Number of Pixel Control Points"],
        "attrs": {
            "long_name": "Longitude\x00",
            "standard_name": "longitude\x00",
            "units": "degrees_east\x00",
            "valid_range": (-180.0, 180.0),
        }
    }
    lat_info = {
        "type": SDC.FLOAT32,
        "data": np.zeros((5, 5), np.float32),
        "dim_labels": ["Number of Scan Lines", "Number of Pixel Control Points"],
        "attrs": {
            "long_name": "Latitude\x00",
            "standard_name": "latitude\x00",
            "units": "degrees_north\x00",
            "valid_range": (-90.0, 90.0),
        }
    }
    _add_variable_to_file(h, "longitude", lon_info)
    _add_variable_to_file(h, "latitude", lat_info)

    chlor_a_info = {
        "type": SDC.FLOAT32,
        "data": np.zeros((5, 5), np.float32),
        "dim_labels": ["Number of Scan Lines", "Number of Pixel Control Points"],
        "attrs": {
            "long_name": "Chlorophyll Concentration, OCI Algorithm\x00",
            "units": "mg m^-3\x00",
            "standard_name": "mass_concentration_of_chlorophyll_in_sea_water\x00",
            "valid_range": (0.001, 100.0),
        }
    }
    _add_variable_to_file(h, "chlor_a", chlor_a_info)
    return [full_path]


def _add_variable_to_file(h, var_name, var_info):
    v = h.create(var_name, var_info['type'], var_info['data'].shape)
    v[:] = var_info['data']
    for dim_count, dimension_name in enumerate(var_info['dim_labels']):
        v.dim(dim_count).setname(dimension_name)
    if var_info.get('fill_value'):
        v.setfillvalue(var_info['fill_value'])
    for attr_key, attr_val in var_info['attrs'].items():
        setattr(v, attr_key, attr_val)


class TestSEADAS:
    """Test the SEADAS L2 file reader."""

    def test_available_reader(self):
        """Test that SEADAS L2 reader is available."""
        assert 'seadas_l2' in available_readers()

    @pytest.mark.parametrize(
        "input_files",
        [
            lazy_fixture("seadas_l2_modis_chlor_a"),
            lazy_fixture("seadas_l2_viirs_npp_chlor_a"),
            lazy_fixture("seadas_l2_viirs_j01_chlor_a"),
        ])
    def test_scene_available_datasets(self, input_files):
        """Test that datasets are available."""
        scene = Scene(reader='seadas_l2', filenames=input_files)
        available_datasets = scene.all_dataset_names()
        assert len(available_datasets) > 0
        assert 'chlor_a' in available_datasets

    @pytest.mark.parametrize(
        ("input_files", "exp_plat", "exp_sensor", "exp_rps"),
        [
            (lazy_fixture("seadas_l2_modis_chlor_a"), "Aqua", {"modis"}, 10),
            (lazy_fixture("seadas_l2_viirs_npp_chlor_a"), "Suomi-NPP", {"viirs"}, 16),
            (lazy_fixture("seadas_l2_viirs_j01_chlor_a"), "NOAA-20", {"viirs"}, 16),
        ])
    def test_load_chlor_a(self, input_files, exp_plat, exp_sensor, exp_rps):
        """Test that we can load 'chlor_a'."""
        scene = Scene(reader='seadas_l2', filenames=input_files)
        scene.load(['chlor_a'])
        data_arr = scene['chlor_a']
        assert data_arr.attrs['platform_name'] == exp_plat
        assert data_arr.attrs['sensor'] == exp_sensor
        assert data_arr.attrs['units'] == 'mg m^-3'
        assert data_arr.dtype.type == np.float32
        assert isinstance(data_arr.attrs["area"], SwathDefinition)
        assert data_arr.attrs["rows_per_scan"] == exp_rps



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'fci_l1c_nc' reader."""

import logging
import os
from unittest import mock

import dask.array as da
import numpy as np
import numpy.testing
import pytest
import xarray as xr

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler


class FakeNetCDF4FileHandler2(FakeNetCDF4FileHandler):
    """Class for faking the NetCDF4 Filehandler."""

    def _get_test_calib_for_channel_ir(self, chroot, meas):
        from pyspectral.blackbody import C_SPEED as c
        from pyspectral.blackbody import H_PLANCK as h
        from pyspectral.blackbody import K_BOLTZMANN as k
        xrda = xr.DataArray
        data = {}
        data[meas + "/radiance_to_bt_conversion_coefficient_wavenumber"] = xrda(955)
        data[meas + "/radiance_to_bt_conversion_coefficient_a"] = xrda(1)
        data[meas + "/radiance_to_bt_conversion_coefficient_b"] = xrda(0.4)
        data[meas + "/radiance_to_bt_conversion_constant_c1"] = xrda(1e11 * 2 * h * c ** 2)
        data[meas + "/radiance_to_bt_conversion_constant_c2"] = xrda(1e2 * h * c / k)
        return data

    def _get_test_calib_for_channel_vis(self, chroot, meas):
        xrda = xr.DataArray
        data = {}
        data["state/celestial/earth_sun_distance"] = xrda(da.repeat(da.array([149597870.7]), 6000))
        data[meas + "/channel_effective_solar_irradiance"] = xrda(50)
        return data

    def _get_test_content_for_channel(self, pat, ch):
        xrda = xr.DataArray
        nrows = 200
        ncols = 11136
        chroot = "data/{:s}"
        meas = chroot + "/measured"
        rad = meas + "/effective_radiance"
        qual = meas + "/pixel_quality"
        index_map = meas + "/index_map"
        rad_conv_coeff = meas + "/radiance_unit_conversion_coefficient"
        pos = meas + "/{:s}_position_{:s}"
        shp = rad + "/shape"
        x = meas + "/x"
        y = meas + "/y"
        data = {}
        ch_str = pat.format(ch)
        ch_path = rad.format(ch_str)

        common_attrs = {
            "scale_factor": 5,
            "add_offset": 10,
            "long_name": "Effective Radiance",
            "units": "mW.m-2.sr-1.(cm-1)-1",
            "ancillary_variables": "pixel_quality"
        }
        if ch == 38:
            fire_line = da.ones((1, ncols), dtype="uint16", chunks=1024) * 5000
            data_without_fires = da.ones((nrows - 1, ncols), dtype="uint16", chunks=1024)
            d = xrda(
                da.concatenate([fire_line, data_without_fires], axis=0),
                dims=("y", "x"),
                attrs={
                    "valid_range": [0, 8191],
                    "warm_scale_factor": 2,
                    "warm_add_offset": -300,
                    **common_attrs
                }
            )
        else:
            d = xrda(
                da.ones((nrows, ncols), dtype="uint16", chunks=1024),
                dims=("y", "x"),
                attrs={
                    "valid_range": [0, 4095],
                    "warm_scale_factor": 1,
                    "warm_add_offset": 0,
                    **common_attrs
                }
            )

        data[ch_path] = d
        data[x.format(ch_str)] = xrda(
            da.arange(1, ncols + 1, dtype="uint16"),
            dims=("x",),
            attrs={
                "scale_factor": -5.58877772833e-05,
                "add_offset": 0.155619515845,
            }
        )
        data[y.format(ch_str)] = xrda(
            da.arange(1, nrows + 1, dtype="uint16"),
            dims=("y",),
            attrs={
                "scale_factor": -5.58877772833e-05,
                "add_offset": 0.155619515845,
            }
        )
        data[qual.format(ch_str)] = xrda(
            da.arange(nrows * ncols, dtype="uint8").reshape(nrows, ncols) % 128,
            dims=("y", "x"))
        # add dummy data for index map starting from 1
        data[index_map.format(ch_str)] = xrda(
            (da.arange(nrows * ncols, dtype="uint16").reshape(nrows, ncols) % 6000) + 1,
            dims=("y", "x"))

        data[rad_conv_coeff.format(ch_str)] = xrda(1234.56)
        data[pos.format(ch_str, "start", "row")] = xrda(0)
        data[pos.format(ch_str, "start", "column")] = xrda(0)
        data[pos.format(ch_str, "end", "row")] = xrda(nrows)
        data[pos.format(ch_str, "end", "column")] = xrda(ncols)
        if pat.startswith("ir") or pat.startswith("wv"):
            data.update(self._get_test_calib_for_channel_ir(chroot.format(ch_str),
                                                            meas.format(ch_str)))
        elif pat.startswith("vis") or pat.startswith("nir"):
            data.update(self._get_test_calib_for_channel_vis(chroot.format(ch_str),
                                                             meas.format(ch_str)))
        data[shp.format(ch_str)] = (nrows, ncols)
        return data

    def _get_test_content_all_channels(self):
        chan_patterns = {
            "vis_{:>02d}": (4, 5, 6, 8, 9),
            "nir_{:>02d}": (13, 16, 22),
            "ir_{:>02d}": (38, 87, 97, 105, 123, 133),
            "wv_{:>02d}": (63, 73),
        }
        data = {}
        for pat in chan_patterns:
            for ch_num in chan_patterns[pat]:
                data.update(self._get_test_content_for_channel(pat, ch_num))
        return data

    def _get_test_content_areadef(self):
        data = {}

        proj = "data/mtg_geos_projection"

        attrs = {
            "sweep_angle_axis": "y",
            "perspective_point_height": "35786400.0",
            "semi_major_axis": "6378137.0",
            "longitude_of_projection_origin": "0.0",
            "inverse_flattening": "298.257223563",
            "units": "m"}
        data[proj] = xr.DataArray(
            0,
            dims=(),
            attrs=attrs)

        # also set attributes cached, as this may be how they are accessed with
        # the NetCDF4FileHandler
        for (k, v) in attrs.items():
            data[proj + "/attr/" + k] = v

        return data

    def _get_test_content_aux_data(self):
        from satpy.readers.fci_l1c_nc import AUX_DATA
        xrda = xr.DataArray
        data = {}
        indices_dim = 6000
        for key, value in AUX_DATA.items():
            # skip population of earth_sun_distance as this is already defined for reflectance calculation
            if key == 'earth_sun_distance':
                continue
            data[value] = xrda(da.arange(indices_dim, dtype="float32"), dims=("index"))

        # compute the last data entry to simulate the FCI caching
        data[list(AUX_DATA.values())[-1]] = data[list(AUX_DATA.values())[-1]].compute()

        return data

    def _get_global_attributes(self):
        data = {}
        attrs = {"platform": "MTI1"}
        for (k, v) in attrs.items():
            data["/attr/" + k] = v
        return data

    def get_test_content(self, filename, filename_info, filetype_info):
        """Get the content of the test data."""
        # mock global attributes
        # - root groups global
        # - other groups global
        # mock data variables
        # mock dimensions
        #
        # ... but only what satpy is using ...

        D = {}
        D.update(self._get_test_content_all_channels())
        D.update(self._get_test_content_areadef())
        D.update(self._get_test_content_aux_data())
        D.update(self._get_global_attributes())
        return D


class FakeNetCDF4FileHandler3(FakeNetCDF4FileHandler2):
    """Mock bad data."""

    def _get_test_calib_for_channel_ir(self, chroot, meas):
        from netCDF4 import default_fillvals
        v = xr.DataArray(default_fillvals["f4"])
        data = {}
        data[meas + "/radiance_to_bt_conversion_coefficient_wavenumber"] = v
        data[meas + "/radiance_to_bt_conversion_coefficient_a"] = v
        data[meas + "/radiance_to_bt_conversion_coefficient_b"] = v
        data[meas + "/radiance_to_bt_conversion_constant_c1"] = v
        data[meas + "/radiance_to_bt_conversion_constant_c2"] = v
        return data

    def _get_test_calib_for_channel_vis(self, chroot, meas):
        data = super()._get_test_calib_for_channel_vis(chroot, meas)
        from netCDF4 import default_fillvals
        v = xr.DataArray(default_fillvals["f4"])
        data[meas + "/channel_effective_solar_irradiance"] = v
        return data


@pytest.fixture
def reader_configs():
    """Return reader configs for FCI."""
    from satpy._config import config_search_paths
    return config_search_paths(
        os.path.join("readers", "fci_l1c_nc.yaml"))


def _get_reader_with_filehandlers(filenames, reader_configs):
    from satpy.readers import load_reader
    reader = load_reader(reader_configs)
    loadables = reader.select_files_from_pathnames(filenames)
    reader.create_filehandlers(loadables)
    return reader


class TestFCIL1cNCReader:
    """Initialize the unittest TestCase for the FCI L1c NetCDF Reader."""

    yaml_file = "fci_l1c_nc.yaml"

    _alt_handler = FakeNetCDF4FileHandler2

    @pytest.fixture(autouse=True, scope="class")
    def fake_handler(self):
        """Wrap NetCDF4 FileHandler with our own fake handler."""
        # implementation strongly inspired by test_viirs_l1b.py
        from satpy.readers.fci_l1c_nc import FCIL1cNCFileHandler
        p = mock.patch.object(
                FCIL1cNCFileHandler,
                "__bases__",
                (self._alt_handler,))
        with p:
            p.is_local = True
            yield p


class TestFCIL1cNCReaderGoodData(TestFCIL1cNCReader):
    """Test FCI L1c NetCDF reader."""

    _alt_handler = FakeNetCDF4FileHandler2

    def test_file_pattern(self, reader_configs):
        """Test file pattern matching."""
        from satpy.readers import load_reader

        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc",
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114442_GTT_DEV_"
            "20170410113934_20170410113942_N__C_0070_0068.nc",
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114451_GTT_DEV_"
            "20170410113942_20170410113951_N__C_0070_0069.nc",
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114500_GTT_DEV_"
            "20170410113951_20170410114000_N__C_0070_0070.nc",
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-TRAIL--L2P-NC4E_C_EUMT_20170410114600_GTT_DEV_"
            "20170410113000_20170410114000_N__C_0070_0071.nc",
        ]

        reader = load_reader(reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        # only 4 out of 5 above should match
        assert len(files) == 4

    _chans = {"solar": ["vis_04", "vis_05", "vis_06", "vis_08", "vis_09",
                        "nir_13", "nir_16", "nir_22"],
              "terran": ["ir_38", "wv_63", "wv_73", "ir_87", "ir_97", "ir_105",
                         "ir_123", "ir_133"]}

    def test_load_counts(self, reader_configs):
        """Test loading with counts."""
        from satpy.tests.utils import make_dataid

        # testing two filenames to test correctly combined
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc",
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114442_GTT_DEV_"
            "20170410113934_20170410113942_N__C_0070_0068.nc",
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)
        res = reader.load(
            [make_dataid(name=name, calibration="counts") for name in
             self._chans["solar"] + self._chans["terran"]], pad_data=False)
        assert 16 == len(res)
        for ch in self._chans["solar"] + self._chans["terran"]:
            assert res[ch].shape == (200 * 2, 11136)
            assert res[ch].dtype == np.uint16
            assert res[ch].attrs["calibration"] == "counts"
            assert res[ch].attrs["units"] == "count"
            if ch == 'ir_38':
                numpy.testing.assert_array_equal(res[ch][~0], 1)
                numpy.testing.assert_array_equal(res[ch][0], 5000)
            else:
                numpy.testing.assert_array_equal(res[ch], 1)

    def test_load_radiance(self, reader_configs):
        """Test loading with radiance."""
        from satpy.tests.utils import make_dataid

        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc",
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)
        res = reader.load(
            [make_dataid(name=name, calibration="radiance") for name in
             self._chans["solar"] + self._chans["terran"]], pad_data=False)
        assert 16 == len(res)
        for ch in self._chans["solar"] + self._chans["terran"]:
            assert res[ch].shape == (200, 11136)
            assert res[ch].dtype == np.float64
            assert res[ch].attrs["calibration"] == "radiance"
            assert res[ch].attrs["units"] == 'mW m-2 sr-1 (cm-1)-1'
            assert res[ch].attrs["radiance_unit_conversion_coefficient"] == 1234.56
            if ch == 'ir_38':
                numpy.testing.assert_array_equal(res[ch][~0], 15)
                numpy.testing.assert_array_equal(res[ch][0], 9700)
            else:
                numpy.testing.assert_array_equal(res[ch], 15)

    def test_load_reflectance(self, reader_configs):
        """Test loading with reflectance."""
        from satpy.tests.utils import make_dataid

        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc",
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)
        res = reader.load(
            [make_dataid(name=name, calibration="reflectance") for name in
             self._chans["solar"]], pad_data=False)
        assert 8 == len(res)
        for ch in self._chans["solar"]:
            assert res[ch].shape == (200, 11136)
            assert res[ch].dtype == np.float64
            assert res[ch].attrs["calibration"] == "reflectance"
            assert res[ch].attrs["units"] == "%"
            numpy.testing.assert_array_almost_equal(res[ch], 100 * 15 * 1 * np.pi / 50)

    def test_load_bt(self, reader_configs, caplog):
        """Test loading with bt."""
        from satpy.tests.utils import make_dataid

        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc",
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)
        with caplog.at_level(logging.WARNING):
            res = reader.load(
                [make_dataid(name=name, calibration="brightness_temperature") for
                 name in self._chans["terran"]], pad_data=False)
            assert caplog.text == ""
        for ch in self._chans["terran"]:
            assert res[ch].shape == (200, 11136)
            assert res[ch].dtype == np.float64
            assert res[ch].attrs["calibration"] == "brightness_temperature"
            assert res[ch].attrs["units"] == "K"

            if ch == 'ir_38':
                numpy.testing.assert_array_almost_equal(res[ch][~0], 209.68274099)
                numpy.testing.assert_array_almost_equal(res[ch][0], 1888.851296)
            else:
                numpy.testing.assert_array_almost_equal(res[ch], 209.68274099)

    def test_load_index_map(self, reader_configs):
        """Test loading of index_map."""
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc"
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)
        res = reader.load(
            [name + '_index_map' for name in
             self._chans["solar"] + self._chans["terran"]], pad_data=False)
        assert 16 == len(res)
        for ch in self._chans["solar"] + self._chans["terran"]:
            assert res[ch + '_index_map'].shape == (200, 11136)
            numpy.testing.assert_array_equal(res[ch + '_index_map'][1, 1], 5138)

    def test_load_aux_data(self, reader_configs):
        """Test loading of auxiliary data."""
        from satpy.readers.fci_l1c_nc import AUX_DATA

        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc"
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)
        res = reader.load(['vis_04_' + key for key in AUX_DATA.keys()],
                          pad_data=False)
        for aux in ['vis_04_' + key for key in AUX_DATA.keys()]:

            assert res[aux].shape == (200, 11136)
            if aux == 'vis_04_earth_sun_distance':
                numpy.testing.assert_array_equal(res[aux][1, 1], 149597870.7)
            else:
                numpy.testing.assert_array_equal(res[aux][1, 1], 5137)

    def test_load_composite(self):
        """Test that composites are loadable."""
        # when dedicated composites for FCI are implemented in satpy,
        # this method should probably move to a dedicated class and module
        # in the tests.compositor_tests package

        from satpy.composites.config_loader import load_compositor_configs_for_sensors
        comps, mods = load_compositor_configs_for_sensors(['fci'])
        assert len(comps["fci"]) > 0
        assert len(mods["fci"]) > 0

    def test_load_quality_only(self, reader_configs):
        """Test that loading quality only works."""
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc",
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)
        res = reader.load(
            [name + '_pixel_quality' for name in
             self._chans["solar"] + self._chans["terran"]], pad_data=False)
        assert 16 == len(res)
        for ch in self._chans["solar"] + self._chans["terran"]:
            assert res[ch + '_pixel_quality'].shape == (200, 11136)
            numpy.testing.assert_array_equal(res[ch + '_pixel_quality'][1, 1], 1)
            assert res[ch + '_pixel_quality'].attrs["name"] == ch + '_pixel_quality'

    def test_platform_name(self, reader_configs):
        """Test that platform name is exposed.

        Test that the FCI reader exposes the platform name.  Corresponds
        to GH issue 1014.
        """
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc",
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)
        res = reader.load(["ir_123"], pad_data=False)
        assert res["ir_123"].attrs["platform_name"] == "MTG-I1"

    def test_excs(self, reader_configs):
        """Test that exceptions are raised where expected."""
        from satpy.tests.utils import make_dataid
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc",
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)

        with pytest.raises(ValueError):
            reader.file_handlers["fci_l1c_fdhsi"][0].get_dataset(make_dataid(name="invalid"), {})
        with pytest.raises(ValueError):
            reader.file_handlers["fci_l1c_fdhsi"][0].get_dataset(
                make_dataid(name="ir_123", calibration="unknown"),
                {"units": "unknown"})

    def test_area_definition_computation(self, reader_configs):
        """Test that the geolocation computation is correct."""
        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc",
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)
        res = reader.load(['ir_105', 'vis_06'], pad_data=False)

        # test that area_ids are harmonisation-conform <platform>_<instrument>_<service>_<resolution>
        assert res['vis_06'].attrs['area'].area_id == 'mtg_fci_fdss_1km'
        assert res['ir_105'].attrs['area'].area_id == 'mtg_fci_fdss_2km'

        area_def = res['ir_105'].attrs['area']
        # test area extents computation
        np.testing.assert_array_almost_equal(np.array(area_def.area_extent),
                                             np.array([-5568062.23065902, 5168057.7600648,
                                                       16704186.692027, 5568062.23065902]))

        # check that the projection is read in properly
        assert area_def.crs.coordinate_operation.method_name == 'Geostationary Satellite (Sweep Y)'
        assert area_def.crs.coordinate_operation.params[0].value == 0.0  # projection origin longitude
        assert area_def.crs.coordinate_operation.params[1].value == 35786400.0  # projection height
        assert area_def.crs.ellipsoid.semi_major_metre == 6378137.0
        assert area_def.crs.ellipsoid.inverse_flattening == 298.257223563
        assert area_def.crs.ellipsoid.is_semi_minor_computed


class TestFCIL1cNCReaderBadData(TestFCIL1cNCReader):
    """Test the FCI L1c NetCDF Reader for bad data input."""

    _alt_handler = FakeNetCDF4FileHandler3

    def test_handling_bad_data_ir(self, reader_configs, caplog):
        """Test handling of bad IR data."""
        from satpy.tests.utils import make_dataid

        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc",
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)
        with caplog.at_level("ERROR"):
            reader.load([make_dataid(
                name="ir_123",
                calibration="brightness_temperature")], pad_data=False)
            assert "cannot produce brightness temperature" in caplog.text

    def test_handling_bad_data_vis(self, reader_configs, caplog):
        """Test handling of bad VIS data."""
        from satpy.tests.utils import make_dataid

        filenames = [
            "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
            "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
            "20170410113925_20170410113934_N__C_0070_0067.nc",
        ]

        reader = _get_reader_with_filehandlers(filenames, reader_configs)
        with caplog.at_level("ERROR"):
            reader.load([make_dataid(
                name="vis_04",
                calibration="reflectance")], pad_data=False)
            assert "cannot produce reflectance" in caplog.text
